"""Runs a pipeline in ./tests/pipelines."""
import logging
from pathlib import Path
import pytest
from unittest.mock import call
import pypyr.pipelinerunner
from tests.common.utils import patch_logger, read_file_to_list

working_dir = Path.joinpath(Path.cwd(), 'tests')


def assert_pipeline_notify_output_is(pipeline_name, expected_notify_output):
    """Assert that the pipeline has the expected output to NOTIFY log.

    Args:
        pipeline_name: str. Name of pipeline to run. Relative to ./tests/
        expected_notify_output: list of str. Entirety of strings expected in
                                the NOTIFY level output during pipeline run.
    """
    with patch_logger('pypyr.steps.echo', logging.NOTIFY) as mock_log:
        pypyr.pipelinerunner.main(pipeline_name=pipeline_name,
                                  pipeline_context_input=None,
                                  working_dir=working_dir)

    assert mock_log.mock_calls == [call(v) for v in expected_notify_output]


def assert_pipeline_notify_match_file(pipeline_name,
                                      expected_notify_output_path):
    """Assert that the pipeline has the expected output to NOTIFY log.

    Args:
        pipeline_name: str. Name of pipeline to run. Relative to ./tests/
        expected_notify_output_path: path-like. Path to text file containing
                                     expected output. Relative to working dir.
    """
    assert_pipeline_notify_output_is(
        pipeline_name,
        read_file_to_list(Path.joinpath(working_dir,
                                        'pipelines',
                                        expected_notify_output_path)))


def assert_pipeline_raises(pipeline_name,
                           expected_error,
                           expected_error_msg,
                           expected_notify_output,
                           context_arg=None):
    """Assert pipeline fails with expected NOTIFY output.

    Will fail if pipeline does NOT raise expected_error.

    Args:
        pipeline_name: str. Name of pipeline to run. Relative to ./tests/
        expected_error: type. Expected exception type.
        expected_error_msg: str. Expected exception msg.
        expected_notify_output: list of str. Entirety of strings expected in
                                the NOTIFY level output during pipeline run.
        context_arg: str. Context argument input.
    """
    with patch_logger('pypyr.steps.echo', logging.NOTIFY) as mock_log:
        with pytest.raises(expected_error) as actual_err:
            pypyr.pipelinerunner.main(pipeline_name=pipeline_name,
                                      pipeline_context_input=context_arg,
                                      working_dir=working_dir)

    if expected_error_msg:
        assert str(actual_err.value) == expected_error_msg

    assert mock_log.mock_calls == [call(v) for v in expected_notify_output]


def assert_pipeline_raises_match_file(pipeline_name,
                                      expected_error,
                                      expected_error_msg,
                                      expected_notify_output_path,
                                      context_arg=None):
    """Assert that the pipeline fails the expected output to NOTIFY log.

    Args:
        pipeline_name: str. Name of pipeline to run. Relative to ./tests/
        expected_notify_output_path: path-like. Path to text file containing
                                     expected output. Relative to working dir.
    """
    assert_pipeline_raises(
        pipeline_name,
        expected_error,
        expected_error_msg,
        read_file_to_list(Path.joinpath(working_dir,
                                        'pipelines',
                                        expected_notify_output_path)),
        context_arg)



"""Control of flow instruction to jump to another step-group."""
import logging
from pypyr.errors import Jump
from pypyr.steps.dsl.cof import control_of_flow_instruction

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Control of flow instruction to jump to another step-group.

    Once jumped, you don't come back here.

    Args:
        context: pypyr.context.Context

    Context expexts the following keys:
        jump: str. Single group-name to run.

    OR

        jump:
            groups: list of str, or str. Single group or list of groups to run.
            success: str. Name of group to run on success completion of groups.
            failure: str. Name of group to run on something going wrong.

    Return:
        None.
    """
    logger.debug("started")

    control_of_flow_instruction(name=__name__,
                                instruction_type=Jump,
                                context=context,
                                context_key='jump')



"""pypyr step that writes payload out to a yaml file."""
import os
import logging
from pypyr.utils.asserts import assert_key_has_value
import pypyr.yaml

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Write payload out to yaml file.

    Args:
        context: pypyr.context.Context. Mandatory.
                 The following context keys expected:
                - fileWriteYaml
                    - path. mandatory. path-like. Write output file to
                      here. Will create directories in path for you.
                    - payload. optional. Write this to output file. If not
                      specified, output entire context.

    Returns:
        None.

    Raises:
        pypyr.errors.KeyNotInContextError: fileWriteYaml or
            fileWriteYaml['path'] missing in context.
        pypyr.errors.KeyInContextHasNoValueError: fileWriteYaml or
            fileWriteYaml['path'] exists but is None.

    """
    logger.debug("started")
    context.assert_key_has_value('fileWriteYaml', __name__)

    input_context = context.get_formatted('fileWriteYaml')
    assert_key_has_value(obj=input_context,
                         key='path',
                         caller=__name__,
                         parent='fileWriteYaml')
    out_path = input_context['path']
    # doing it like this to safeguard against accidentally dumping all context
    # with potentially sensitive values in it to disk if payload exists but is
    # None.
    is_payload_specified = 'payload' in input_context

    yaml_writer = pypyr.yaml.get_yaml_parser_roundtrip_for_context()

    logger.debug("opening destination file for writing: %s", out_path)
    os.makedirs(os.path.abspath(os.path.dirname(out_path)), exist_ok=True)
    with open(out_path, 'w') as outfile:
        if is_payload_specified:
            payload = input_context['payload']
        else:
            payload = context.get_formatted_value(context)

        yaml_writer.dump(payload, outfile)

    logger.info("formatted context content and wrote to %s", out_path)
    logger.debug("done")



"""pypyr step that parses yaml for string substitutions, writes output."""
import logging
from pypyr.steps.dsl.fileinoutrewriter import ObjectRewriterStep
from pypyr.utils.filesystem import YamlRepresenter

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Parse input yaml file and substitute {tokens} from context.

    Loads yaml into memory to do parsing, so be aware of big files.

    Args:
        context: pypyr.context.Context. Mandatory.
        - fileFormatYaml
            - in. mandatory.
              str, path-like, or an iterable (list/tuple) of
              strings/paths. Each str/path can be a glob, relative or
              absolute path.
            - out. optional. path-like.
              Can refer to a file or a directory.
              will create directory structure if it doesn't exist. If
              in-path refers to >1 file (e.g it's a glob or list), out
              path can only be a directory - it doesn't make sense to
              write >1 file to the same single file (this is not an
              appender.) To ensure out_path is read as a directory and
              not a file, be sure to have the path separator (/) at the
              end.
              If out_path is not specified or None, will in-place edit
              and overwrite the in-files.

    Returns:
        None.

    Raises:
        FileNotFoundError: take a guess
        pypyr.errors.KeyNotInContextError: fileFormatYaml or
            fileFormatYaml['in'] missing in context.
        pypyr.errors.KeyInContextHasNoValueError: fileFormatYaml or
            fileFormatYaml['in'] exists but is None.

    """
    logger.debug("started")

    ObjectRewriterStep(__name__, 'fileFormatYaml', context).run_step(
        YamlRepresenter())

    logger.debug("done")



"""pypyr step that writes payload out to a json file."""
import json
import logging
import os
from pypyr.utils.asserts import assert_key_has_value
# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Write payload out to json file.

    Args:
        context: pypyr.context.Context. Mandatory.
                 The following context keys expected:
                - fileWriteJson
                    - path. mandatory. path-like. Write output file to
                      here. Will create directories in path for you.
                    - payload. optional. Write this key to output file. If not
                      specified, output entire context.

    Returns:
        None.

    Raises:
        pypyr.errors.KeyNotInContextError: fileWriteJson or
            fileWriteJson['path'] missing in context.
        pypyr.errors.KeyInContextHasNoValueError: fileWriteJson or
            fileWriteJson['path'] exists but is None.

    """
    logger.debug("started")
    context.assert_key_has_value('fileWriteJson', __name__)

    input_context = context.get_formatted('fileWriteJson')
    assert_key_has_value(obj=input_context,
                         key='path',
                         caller=__name__,
                         parent='fileWriteJson')

    out_path = input_context['path']
    # doing it like this to safeguard against accidentally dumping all context
    # with potentially sensitive values in it to disk if payload exists but is
    # None.
    is_payload_specified = 'payload' in input_context

    logger.debug("opening destination file for writing: %s", out_path)
    os.makedirs(os.path.abspath(os.path.dirname(out_path)), exist_ok=True)
    with open(out_path, 'w') as outfile:
        if is_payload_specified:
            payload = input_context['payload']
        else:
            payload = context.get_formatted_value(context)

        json.dump(payload, outfile,
                  indent=2, ensure_ascii=False)

    logger.info("formatted context content and wrote to %s", out_path)
    logger.debug("done")



"""Step that gets $ENVs, using a default value if the $ENV doesn't exist."""
import os
import logging
from pypyr.errors import ContextError, KeyNotInContextError

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Get $ENVs, allowing a default if not found.

    Set context properties from environment variables, and specify a default
    if the environment variable is not found.

    This differs from pypyr.steps.env get, which raises an error if attempting
    to read an $ENV that doesn't exist.

    Args:
        context. mandatory. Context is a pypyr Context.

    Input context is:
        envGet:
            - env: 'envvarnamehere'
              key: 'savetocontexthere'
              default: 'save this to key if env doesnt exist'

    'env' is the bare environment variable name, do not put the $ in front of
    it.

    Will process as many env/key/default pairs as exist in the list under
    envGet.

    Returns:
        None.

    Raises:
        ContextError: envGet is not a list of dicts.
        KeyNotInContextError: envGet env or key doesn't exist.

    """
    logger.debug("started")
    assert context, f"context must have value for {__name__}"

    context.assert_key_has_value('envGet', __name__)

    get_items = context.get_formatted('envGet')
    # allow a list OR a single getenv dict
    if isinstance(context['envGet'], dict):
        get_items = [get_items]

    get_count = 0

    for get_me in get_items:
        (env, key, has_default, default) = get_args(get_me)

        logger.debug("setting context %s to $ENV %s", key, env)

        if env in os.environ:
            context[key] = os.environ[env]
            get_count += 1
        else:
            logger.debug("$ENV %s not found.", env)
            if has_default:
                logger.debug("Using default value for %s instead.", env)
                context[key] = os.environ.get(env, default)
                get_count += 1
            else:
                logger.debug(
                    "No default value for %s found. Doin' nuthin'.", env)

    logger.info("saved %s $ENVs to context.", get_count)


def get_args(get_item):
    """Parse env, key, default out of input dict.

    Args:
        get_item: dict. contains keys env/key/default

    Returns:
        (env, key, has_default, default) tuple, where
            env: str. env var name.
            key: str. save env value to this context key.
            has_default: bool. True if default specified.
            default: the value of default, if specified.

    Raises:
        ContextError: envGet is not a list of dicts.
        KeyNotInContextError: If env or key not found in get_config.

    """
    if not isinstance(get_item, dict):
        raise ContextError('envGet must contain a list of dicts.')

    env = get_item.get('env', None)

    if not env:
        raise KeyNotInContextError(
            'context envGet[env] must exist in context for envGet.')

    key = get_item.get('key', None)

    if not key:
        raise KeyNotInContextError(
            'context envGet[key] must exist in context for envGet.')

    if 'default' in get_item:
        has_default = True
        default = get_item['default']
    else:
        has_default = False
        default = None

    return (env, key, has_default, default)



"""pypyr pipeline runner.

Runs the pipeline specified by the input pipeline_name parameter.
Pipelines must have a "steps" list-like attribute.

Attributes:
    pipeline_cache: global instance of the pipeline yaml cache.
                    Use this attribute to access the cache from elsewhere.
"""
import logging
import pypyr.context
import pypyr.log.logger
import pypyr.moduleloader
from pypyr.cache.parsercache import contextparser_cache
from pypyr.errors import Stop, StopPipeline, StopStepGroup
from pypyr.cache.pipelinecache import pipeline_cache
from pypyr.stepsrunner import StepsRunner
import pypyr.yaml

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)


def get_parsed_context(pipeline, context_in_args):
    """Execute get_parsed_context handler if specified.

    Dynamically load the module specified by the context_parser key in pipeline
    dict and execute the get_parsed_context function on that module.

    Args:
        pipeline: dict. Pipeline object.
        context_in_args: list of string. Input arguments from console.

    Returns:
        pypyr.context.Context() instance.

    Raises:
        AttributeError: parser specified on pipeline missing get_parsed_context
                        function.

    """
    logger.debug("starting")

    if 'context_parser' in pipeline:
        parser_module_name = pipeline['context_parser']
        logger.debug("context parser specified: %s", parser_module_name)
        get_parsed_context = contextparser_cache.get_context_parser(
            parser_module_name)

        logger.debug("running parser %s", parser_module_name)
        result_context = get_parsed_context(context_in_args)
        logger.debug("context parse %s done", parser_module_name)
        # Downstream steps likely to expect context not to be None, hence
        # empty rather than None.
        if result_context is None:
            logger.debug(
                "%s returned None. Using empty context instead",
                parser_module_name
            )
            return pypyr.context.Context()
        else:
            return pypyr.context.Context(result_context)
    else:
        logger.debug("pipeline does not have custom context parser. Using "
                     "empty context.")
        logger.debug("done")
        # initialize to an empty dictionary because you want to be able to run
        # with no context.
        return pypyr.context.Context()


def main(
    pipeline_name,
    pipeline_context_input,
    working_dir,
    groups=None,
    success_group=None,
    failure_group=None
):
    """Entry point for pypyr pipeline runner.

    Call this once per pypyr run. Call me if you want to run a pypyr pipeline
    from your own code. This function does some one-off 1st time initialization
    before running the actual pipeline.

    If you're invoking pypyr from your own application via the API,
    it's your responsibility to set up and configure logging. If you just want
    to replicate the log handlers & formatters that the pypyr cli uses, you can
    call pypyr.log.logger.set_root_logger() before invoking this function
    (pipelinerunner.main())

    Be aware that if you invoke this method, pypyr adds a NOTIFY - 25 custom
    log-level and notify() function to logging.

    pipeline_name.yaml should be in the working_dir/pipelines/ directory.

    Args:
        pipeline_name: string. Name of pipeline, sans .yaml at end.
        pipeline_context_input: string. Initialize the pypyr context with this
                                string.
        working_dir: path. looks for ./pipelines and modules in this directory.
        groups: list of str. step-group names to run in pipeline.
        success_group: str. step-group name to run on success completion.
        failure_group: str. step-group name to run on pipeline failure.

    Returns:
        None

    """
    pypyr.log.logger.set_up_notify_log_level()

    logger.debug("starting pypyr")

    # pipelines specify steps in python modules that load dynamically.
    # make it easy for the operator so that the cwd is automatically included
    # without needing to pip install a package 1st.
    pypyr.moduleloader.set_working_directory(working_dir)

    try:
        load_and_run_pipeline(pipeline_name=pipeline_name,
                              pipeline_context_input=pipeline_context_input,
                              groups=groups,
                              success_group=success_group,
                              failure_group=failure_group)
    except Stop:
        logger.debug("Stop: stopped pypyr")

    logger.debug("pypyr done")


def prepare_context(pipeline, context_in_args, context):
    """Prepare context for pipeline run.

    Args:
        pipeline: dict. Dictionary representing the pipeline.
        context_in_args: list of str. Args used to initialize context.
        context: pypyr.context.Context. Merge any new context generated from
                 context_in_string into this context instance.

    Returns:
        None. The context instance to use for the pipeline run is contained
              in the context arg, it's not passed back as a function return.

    """
    logger.debug("starting")

    parsed_context = get_parsed_context(
        pipeline=pipeline,
        context_in_args=context_in_args)

    context.update(parsed_context)

    logger.debug("done")


def load_and_run_pipeline(pipeline_name,
                          pipeline_context_input=None,
                          context=None,
                          parse_input=True,
                          loader=None,
                          groups=None,
                          success_group=None,
                          failure_group=None):
    """Load and run the specified pypyr pipeline.

    This function runs the actual pipeline by name. If you are running another
    pipeline from within a pipeline, call this, not main(). Do call main()
    instead for your 1st pipeline if there are pipelines calling pipelines.

    By default pypyr uses file loader. This means that pipeline_name.yaml
    should be in the working_dir/ directory if you're using fileloader.

    Look for pipelines and modules in the working_dir. Set the working_dir by
    calling pypyr.moduleloader.set_working_directory('/my/dir')

    Args:
        pipeline_name (str): Name of pipeline, sans .yaml at end.
        pipeline_context_input (str): Initialize the pypyr context with this
                                 string.
        context (pypyr.context.Context): Use if you already have a
                 Context object, such as if you are running a pipeline from
                 within a pipeline and you want to re-use the same context
                 object for the child pipeline. Any mutations of the context by
                 the pipeline will be against this instance of it.
        parse_input (bool): run context_parser in pipeline.
        loader (str): str. optional. Absolute name of pipeline loader module.
                If not specified will use pypyr.pypeloaders.fileloader.
        groups: list of str. step-group names to run in pipeline.
        success_group: str. step-group name to run on success completion.
        failure_group: str. step-group name to run on pipeline failure.

    Returns:
        None

    """
    logger.debug("you asked to run pipeline: %s", pipeline_name)

    logger.debug("you set the initial context to: %s", pipeline_context_input)

    if context is None:
        context = pypyr.context.Context()
        context.pipeline_name = pipeline_name
        context.working_dir = pypyr.moduleloader.get_working_directory()

    # pipeline loading deliberately outside of try catch. The try catch will
    # try to run a failure-handler from the pipeline, but if the pipeline
    # doesn't exist there is no failure handler that can possibly run so this
    # is very much a fatal stop error.
    pipeline_definition = pipeline_cache.get_pipeline(
        pipeline_name=pipeline_name,
        loader=loader)

    run_pipeline(
        pipeline=pipeline_definition,
        pipeline_context_input=pipeline_context_input,
        context=context,
        parse_input=parse_input,
        groups=groups,
        success_group=success_group,
        failure_group=failure_group
    )


def run_pipeline(pipeline,
                 context,
                 pipeline_context_input=None,
                 parse_input=True,
                 groups=None,
                 success_group=None,
                 failure_group=None):
    """Run the specified pypyr pipeline.

    This function runs the actual pipeline. If you are running another
    pipeline from within a pipeline don't call main(). Do call main()
    instead for your 1st pipeline, if there are subsequent pipelines calling
    pipelines use load_and_run_pipeline or run_pipeline.

    Pipeline and context should be already loaded. If pipeline not loaded yet,
    you probably want to call load_and_run_pipeline instead.

    If none of groups, success_group & failure_group specified, defaults to
    ['steps'], on_success, on_failure. If any of groups, success_group or
    failure_group specified, will ONLY run the specified (i.e if you specify
    groups you don't get on_success/on_failure groups unless you specify these
    explicitly.)

    Args:
        pipeline (dict): Dictionary representing the pipeline.
        context (pypyr.context.Context): Reusable context object.
        pipeline_context_input (str): Initialize the pypyr context with this
                                string.
        parse_input (bool): run context_parser in pipeline.
        groups: list of str. step-group names to run in pipeline.
        success_group: str. step-group name to run on success completion.
        failure_group: str. step-group name to run on pipeline failure.

    Returns:
        None

    """
    logger.debug("starting")

    if not groups:
        groups = ['steps']

        if not success_group and not failure_group:
            success_group = 'on_success'
            failure_group = 'on_failure'

    steps_runner = StepsRunner(pipeline_definition=pipeline, context=context)

    try:
        if parse_input:
            logger.debug("executing context_parser")
            prepare_context(pipeline=pipeline,
                            context_in_args=pipeline_context_input,
                            context=context)
        else:
            logger.debug("skipping context_parser")
    except Exception:
        # yes, yes, don't catch Exception. Have to, though, to run the failure
        # handler. Also, it does raise it back up.
        logger.error("Something went wrong. Will now try to run %s",
                     failure_group)

        # failure_step_group will log but swallow any errors except Stop
        try:
            steps_runner.run_failure_step_group(failure_group)
        except StopStepGroup:
            pass
        except Stop:
            raise

        logger.debug("Raising original exception to caller.")
        raise

    try:
        steps_runner.run_step_groups(groups=groups,
                                     success_group=success_group,
                                     failure_group=failure_group)
    except StopPipeline:
        logger.debug("StopPipeline: stopped %s", context.pipeline_name)



"""Step that echos pypyr version."""
import logging
import pypyr.version

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Output pypyr version in format 'pypyr x.y.z python a.b.c'."""
    logger.debug("started")

    logger.notify("pypyr version is: %s", pypyr.version.get_version())

    logger.debug("done")



"""pypyr step that loads yaml file into context."""
from collections.abc import MutableMapping
import logging
import ruamel.yaml as yaml
from pypyr.utils.asserts import assert_key_has_value
# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Load a yaml file into the pypyr context.

    Yaml parsed from the file will be merged into the pypyr context. This will
    overwrite existing values if the same keys are already in there.
    I.e if file yaml has {'eggs' : 'boiled'} and context {'eggs': 'fried'}
    already exists, returned context['eggs'] will be 'boiled'.

    Args:
        context: pypyr.context.Context. Mandatory.
                 The following context key must exist
                - fetchYaml
                    - path. path-like. Path to file on disk.
                    - key. string. If exists, write yaml to this context key.
                      Else yaml writes to context root.

    All inputs support formatting expressions.

    Also supports a passing path as string to fetchYaml, but in this case you
    won't be able to specify a key.

    Returns:
        None. updates context arg.

    Raises:
        FileNotFoundError: take a guess
        pypyr.errors.KeyNotInContextError: fetchYamlPath missing in context.
        pypyr.errors.KeyInContextHasNoValueError: fetchYamlPath exists but is
                                                  None.

    """
    logger.debug("started")

    context.assert_key_has_value(key='fetchYaml', caller=__name__)

    fetch_yaml_input = context.get_formatted('fetchYaml')

    if isinstance(fetch_yaml_input, str):
        file_path = fetch_yaml_input
        destination_key = None
    else:
        assert_key_has_value(obj=fetch_yaml_input,
                             key='path',
                             caller=__name__,
                             parent='fetchYaml')
        file_path = fetch_yaml_input['path']
        destination_key = fetch_yaml_input.get('key', None)

    logger.debug("attempting to open file: %s", file_path)
    with open(file_path) as yaml_file:
        yaml_loader = yaml.YAML(typ='safe', pure=True)
        payload = yaml_loader.load(yaml_file)

    if destination_key:
        logger.debug("yaml file loaded. Writing to context %s",
                     destination_key)
        context[destination_key] = payload
    else:
        if not isinstance(payload, MutableMapping):
            raise TypeError(
                "yaml input should describe a dictionary at the top "
                "level when fetchYamlKey isn't specified. You should have "
                "something like \n'key1: value1'\n key2: value2'\n"
                "in the yaml top-level, not \n'- value1\n - value2'")

        logger.debug("yaml file loaded. Merging into pypyr context. . .")
        context.update(payload)

    logger.info("yaml file written into pypyr context. Count: %s",
                len(payload))
    logger.debug("done")



"""Archive and extract tars."""
import logging
import tarfile
from pypyr.errors import KeyNotInContextError

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Archive and/or extract tars with or without compression.

    Args:
        context: dictionary-like. Mandatory.

        Expects the following context:
        tar:
            extract:
                - in: /path/my.tar
                  out: /out/path
            archive:
                - in: /dir/to/archive
                  out: /out/destination.tar
            format: ''

        tar['format'] - if not specified, defaults to lzma/xz
                       Available options:
                        - '' - no compression
                        - gz (gzip)
                        - bz2 (bzip2)
                        - xz (lzma)

    This step will run whatever combination of Extract and Archive you specify.
    Regardless of combination, execution order is Extract, Archive.

    Source and destination paths support {key} string interpolation.

    Never extract archives from untrusted sources without prior inspection.
    It is possible that files are created outside of path, e.g. members that
    have absolute filenames starting with "/" or filenames with two dots "..".
    """
    logger.debug("started")

    assert context, f"context must have value for {__name__}"

    found_at_least_one = False

    context.assert_key_has_value('tar', __name__)

    tar_context = context.get_formatted('tar')

    if tar_context.get('extract', None):
        found_at_least_one = True
        tar_extract(tar_context)

    if tar_context.get('archive', None):
        found_at_least_one = True
        tar_archive(tar_context)

    if not found_at_least_one:
        # This will raise exception on first item with a problem.
        raise KeyNotInContextError('pypyr.steps.tar must have either extract '
                                   'or archive specified under the tar key. '
                                   'Or both of these. It has neither.')

    logger.debug("done")


def get_file_mode_for_reading(context_tar):
    """Get file mode for reading from tar['format'].

    This should return r:*, r:gz, r:bz2 or r:xz. If user specified something
    wacky in tar.Format, that's their business.

    In theory r:* will auto-deduce the correct format.
    """
    format = context_tar.get('format', None)

    if format or format == '':
        mode = f"r:{format}"
    else:
        mode = 'r:*'

    return mode


def get_file_mode_for_writing(context_tar):
    """Get file mode for writing from tar['format'].

    This should return w:, w:gz, w:bz2 or w:xz. If user specified something
    wacky in tar.Format, that's their business.
    """
    format = context_tar.get('format', None)
    # slightly weird double-check because falsy format could mean either format
    # doesn't exist in input, OR that it exists and is empty. Exists-but-empty
    # has special meaning - default to no compression.
    if format or format == '':
        mode = f"w:{format}"
    else:
        mode = 'w:xz'

    return mode


def tar_archive(context_tar):
    """Archive specified path to a tar archive.

    Args:
        context_tar: dictionary-like. context is mandatory.
            context['tar']['archive'] must exist. It's a dictionary.
            keys are the paths to archive.
            values are the destination output paths.

    Example:
        tar:
            archive:
                - in: path/to/dir
                  out: path/to/destination.tar.xs
                - in: another/my.file
                  out: ./my.tar.xs

        This will archive directory path/to/dir to path/to/destination.tar.xs,
        and also archive file another/my.file to ./my.tar.xs
    """
    logger.debug("start")

    mode = get_file_mode_for_writing(context_tar)

    for item in context_tar['archive']:
        # value is the destination tar. Allow string interpolation.
        destination = item['out']
        # key is the source to archive
        source = item['in']
        with tarfile.open(destination, mode) as archive_me:
            logger.debug("Archiving '%s' to '%s'", source, destination)

            archive_me.add(source, arcname='.')
            logger.info("Archived '%s' to '%s'", source, destination)

    logger.debug("end")


def tar_extract(context_tar):
    """Extract all members of tar archive to specified path.

    Args:
        context_tar: dictionary-like. context is mandatory.
            context['tar']['extract'] must exist. It's a dictionary.
            keys are the path to the tar to extract.
            values are the destination paths.

    Example:
        tar:
            extract:
                - in: path/to/my.tar.xs
                  out: /path/extract/here
                - in: another/tar.xs
                  out: .

        This will extract path/to/my.tar.xs to /path/extract/here, and also
        extract another/tar.xs to $PWD.
    """
    logger.debug("start")

    mode = get_file_mode_for_reading(context_tar)

    for item in context_tar['extract']:
        # in is the path to the tar to extract. Allows string interpolation.
        source = item['in']
        # out is the outdir, dhur. Allows string interpolation.
        destination = item['out']
        with tarfile.open(source, mode) as extract_me:
            logger.debug("Extracting '%s' to '%s'", source, destination)

            extract_me.extractall(destination)
            logger.info("Extracted '%s' to '%s'", source, destination)

    logger.debug("end")



"""Custom exceptions for pypyr.

All pypyr specific exceptions derive from Error.
"""


def get_error_name(error):
    """Return canonical error name as string.

    For builtin errors like ValueError or Exception, will return the bare
    name, like ValueError or Exception.

    For all other exceptions, will return modulename.errorname, such as
    arbpackage.mod.myerror

    Args:
        error: Exception object.

    Returns:
        str. Canonical error name.

    """
    error_type = type(error)
    if error_type.__module__ in ['__main__', 'builtins']:
        return error_type.__name__
    else:
        return f'{error_type.__module__}.{error_type.__name__}'


class Error(Exception):
    """Base class for all pypyr exceptions."""


class ContextError(Error):
    """Error in the pypyr context."""


class HandledError(Error):
    """Error that has already been saved to errors context collection."""


class KeyInContextHasNoValueError(ContextError):
    """pypyr context[key] doesn't have a value."""


class KeyNotInContextError(ContextError, KeyError):
    """Key not found in the pypyr context."""

    def __str__(self):
        """Avoid KeyError custom error formatting."""
        return super(Exception, self).__str__()


class LoopMaxExhaustedError(Error):
    """Max attempts reached during looping."""


class PipelineDefinitionError(Error):
    """Pipeline definition incorrect. Likely a yaml error."""


class PipelineNotFoundError(Error):
    """Pipeline not found in working dir or in pypyr install dir."""


class PlugInError(Error):
    """Pypyr plug - ins should sub - class this."""


class PyModuleNotFoundError(Error, ModuleNotFoundError):
    """Could not load python module because it wasn't found."""


# -------------------------- Control of Flow Instructions ---------------------
class Stop(Error):
    """Control of flow. Stop all execution."""


class StopPipeline(Stop):
    """Control of flow. Stop current pipeline execution."""


class StopStepGroup(Stop):
    """Control of flow. Stop current step - group execution."""


class ControlOfFlowInstruction(Error):
    """Control of flow instructions should inherit from this.

    Attributes:
        groups: list of str. List of step - groups to execute.
        success_group: str. Step - group to execute on success condition.
        failure_group: str. Step - group to execute on failure condition.
        original_config: tuple. (key, value). Key-Value pair of context dict
                        item used as input on the step that created this
                        ControlOfFlowInstruction.
    """

    def __init__(self, groups, success_group, failure_group, original_config):
        """Initialize the control of flow instruction.

        Args:
            groups: list of str. List of step - groups to execute.
            success_group: str. Step - group to execute on success condition.
            failure_group: str. Step - group to execute on failure condition.
            original_config: tuple. (key, value). Key/Value pair of Context
                             input for the control-of-flow step.
        """
        self.groups = groups
        self.success_group = success_group
        self.failure_group = failure_group
        self.original_config = original_config


class Call(ControlOfFlowInstruction):
    """Stop current step, call a new step group, resume current step after."""


class Jump(ControlOfFlowInstruction):
    """Stop step execution and jump to a new step group."""

# -------------------------- END Control of Flow Instructions -----------------



"""pypyr dynamic modules and path discovery.

Load modules dynamically, find things on file-system.
"""

import importlib
import logging
from pathlib import Path
import sys
from pypyr.errors import PyModuleNotFoundError

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)


class WorkingDir():
    """The Working Directory.

    Attributes:
        cwd. Path-like. The working directory.
    """

    def __init__(self, cwd=None):
        """Initialize cwd to input or default to current."""
        self.cwd = cwd if cwd else Path.cwd()


_working_dir = WorkingDir()


def get_module(module_abs_import):
    """Use importlib to get the module dynamically.

    Get instance of the module specified by the module_abs_import.
    This means that module_abs_import must be resolvable from this package.

    Args:
        module_abs_import: string. Absolute name of module to import.

    Raises:
        PyModuleNotFoundError: if module not found.

    """
    logger.debug("starting")
    logger.debug("loading module %s", module_abs_import)
    try:
        imported_module = importlib.import_module(module_abs_import)
        logger.debug("done")
        return imported_module
    except ModuleNotFoundError as err:
        if err.name != module_abs_import:
            extended_msg = (
                f'error importing module {err.name} in {module_abs_import}')
            logger.error("Couldn't import module %s in this module: %s",
                         err.name,
                         module_abs_import)
        else:
            extended_msg = (
                f"{module_abs_import}.py should be in your working "
                "dir or it should be installed to the python path."
                "\nIf you have 'package.sub.mod' your current working "
                "dir should contain ./package/sub/mod.py\n"
                "If you specified 'mymodulename', your current "
                "working dir should contain ./mymodulename.py\n"
                "If the module is not in your current working dir, it "
                "must exist in your current python path - so you "
                "should have run pip install or setup.py")
            logger.error("The module doesn't exist. "
                         "Looking for a file like this: %s",
                         module_abs_import)
        raise PyModuleNotFoundError(extended_msg) from err


def get_working_directory():
    """Return current working directory as Path."""
    return _working_dir.cwd


def set_working_directory(working_directory):
    """Add working_directory to sys.paths.

    This allows dynamic loading of arbitrary python modules in cwd.

    Args:
        working_directory: string. path to add to sys.paths

    """
    logger.debug("starting")

    logger.debug("adding %s to sys.paths", working_directory)
    # sys path doesn't accept Path
    sys.path.append(str(working_directory))
    _working_dir.cwd = working_directory

    logger.debug("done")



"""pypyr step that parses a json string to context."""
from collections.abc import MutableMapping
import json
import logging
from pypyr.errors import KeyInContextHasNoValueError
from pypyr.utils.asserts import assert_key_exists
# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Parse input string into Context as an object.

    Args:
        context: pypyr.context.Context. Mandatory.
                 The following context keys expected:
                - jsonParse
                    - json. string or formatting expression evaluating to a
                      string of json.
                    - key. string. If exists, write json structure to this
                      context key. Else json writes to context root.

    Returns:
        None.

    Raises:
        pypyr.errors.KeyNotInContextError: jsonParse or jsonParse.json missing
                                           in context.
        pypyr.errors.KeyInContextHasNoValueError: jsonParse.json exists but is
                                                  empty.

    """
    logger.debug("started")
    context.assert_key_has_value('jsonParse', __name__)

    input_context = context.get_formatted('jsonParse')
    assert_key_exists(obj=input_context,
                      key='json',
                      caller=__name__,
                      parent='jsonParse')

    destination_key = input_context.get('key', None)

    json_string = input_context['json']
    if not json_string:
        raise KeyInContextHasNoValueError(
            'jsonParse.json exists but is empty. It should be a valid json '
            'string for pypyr.steps.jsonparse. For example: '
            '\'{"key1": "value1", "key2": "value2"}\'')

    payload = json.loads(json_string)

    if destination_key:
        logger.debug("json string parsed. Writing to context %s",
                     destination_key)
        context[destination_key] = payload
    else:
        if not isinstance(payload, MutableMapping):
            raise TypeError(
                'json input should describe an object at the top '
                'level when jsonParse.key isn\'t specified. You should have '
                'something like \'{"key1": "value1", "key2": "value2"}\' '
                'in the json top-level, not ["value1", "value2"]')

        logger.debug("json string parsed. Merging into pypyr context at %s...",
                     destination_key)
        context.update(payload)

    logger.info("json string parsed into pypyr context.")

    logger.debug("done")



"""pypyr step yaml definition for commands - domain specific language."""
import shlex
import subprocess
import logging
from pypyr.errors import ContextError
from pypyr.utils import types

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


class CmdStep():
    """A pypyr step that represents a command runner step.

    This models a step that takes config like this:
        cmd: <<cmd string>>

        OR, as a dict
        cmd:
            run: str. mandatory. command + args to execute.
            save: bool. defaults False. save output to cmdOut.

    If save is True, will save the output to context as follows:
        cmdOut:
            returncode: 0
            stdout: 'stdout str here. None if empty.'
            stderr: 'stderr str here. None if empty.'

    cmdOut.returncode is the exit status of the called process. Typically 0
    means OK. A negative value -N indicates that the child was terminated by
    signal N (POSIX only).

    The run_step method does the actual work. init loads the yaml.
    """

    def __init__(self, name, context):
        """Initialize the CmdStep.

        The step config in the context dict looks like this:
            cmd: <<cmd string>>

            OR, as a dict
            cmd:
                run: str. mandatory. command + args to execute.
                save: bool. optional. defaults False. save output to cmdOut.
                cwd: str/path. optional. if specified, change the working
                     directory just for the duration of the command.

        Args:
            name: Unique name for step. Likely __name__ of calling step.
            context: pypyr.context.Context. Look for config in this context
                     instance.

        """
        assert name, ("name parameter must exist for CmdStep.")
        assert context, ("context param must exist for CmdStep.")
        # this way, logs output as the calling step, which makes more sense
        # to end-user than a mystery steps.dsl.blah logging output.
        self.logger = logging.getLogger(name)

        context.assert_key_has_value(key='cmd', caller=name)

        self.context = context
        self.is_save = False

        cmd_config = context.get_formatted('cmd')

        if isinstance(cmd_config, str):
            self.cmd_text = cmd_config
            self.cwd = None
            self.logger.debug("Processing command string: %s", cmd_config)
        elif isinstance(cmd_config, dict):
            context.assert_child_key_has_value(parent='cmd',
                                               child='run',
                                               caller=name)

            self.cmd_text = cmd_config['run']
            self.is_save = types.cast_to_bool(cmd_config.get('save', False))

            cwd_string = cmd_config.get('cwd', None)
            if cwd_string:
                self.cwd = cwd_string
                self.logger.debug("Processing command string in dir "
                                  "%s: %s", self.cwd, self.cmd_text)
            else:
                self.cwd = None
                self.logger.debug("Processing command string: %s",
                                  self.cmd_text)

        else:
            raise ContextError(f"{name} cmd config should be either a simple "
                               "string cmd='mycommandhere' or a dictionary "
                               "cmd={'run': 'mycommandhere', 'save': False}.")

    def run_step(self, is_shell):
        """Run a command.

        Runs a program or executable. If is_shell is True, executes the command
        through the shell.

        Args:
            is_shell: bool. defaults False. Set to true to execute cmd through
                      the default shell.
        """
        assert is_shell is not None, ("is_shell param must exist for CmdStep.")

        # why? If shell is True, it is recommended to pass args as a string
        # rather than as a sequence.
        if is_shell:
            args = self.cmd_text
        else:
            args = shlex.split(self.cmd_text)

        if self.is_save:
            completed_process = subprocess.run(args,
                                               cwd=self.cwd,
                                               shell=is_shell,
                                               # capture_output=True,only>py3.7
                                               stdout=subprocess.PIPE,
                                               stderr=subprocess.PIPE,
                                               # text=True, only>=py3.7,
                                               universal_newlines=True)
            self.context['cmdOut'] = {
                'returncode': completed_process.returncode,
                'stdout': (completed_process.stdout.rstrip()
                           if completed_process.stdout else None),
                'stderr': (completed_process.stderr.rstrip()
                           if completed_process.stderr else None)
            }

            # when capture is true, output doesn't write to stdout
            self.logger.info("stdout: %s", completed_process.stdout)
            if completed_process.stderr:
                self.logger.error("stderr: %s", completed_process.stderr)

            # don't swallow the error, because it's the Step swallow decorator
            # responsibility to decide to ignore or not.
            completed_process.check_returncode()
        else:
            # check=True throws CalledProcessError if exit code != 0
            subprocess.run(args, shell=is_shell, check=True, cwd=self.cwd)



"""pypyr step that sets default values in context.

Sets values in context if they do not exist already. Does not overwrite
existing values. Supports nested hierarchies.

Example:
Given a context like this:
    key1: value1
    key2:
        key2.1: value2.1
    key3: None

And defaults input like this:
    key1: 'updated value here won't overwrite since it already exists'
    key2:
        key2.2: value2.2
    key3: 'key 3 exists so I won't overwrite

Will result in context:
    key1: value1
    key2:
        key2.1: value2.1
        key2.2: value2.2
    key3: None

By comparison, the in step decorator, contextcopy, contextsetf and contextmerge
overwrite values that are in context already.

Applies string interpolation as it adds. String interpolation applies to keys
and values.
"""
import logging

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Set hierarchy into context with substitutions if it doesn't exist yet.

    context is a dictionary or dictionary-like.
    context['defaults'] must exist. It's a dictionary.

    Will iterate context['defaults'] and add these as new values where
    their keys don't already exist. While it's doing so, it will leave
    all other values in the existing hierarchy untouched.

    List merging is purely additive, with no checks for uniqueness or already
    existing list items. E.g context [0,1,2] with contextMerge=[2,3,4]
    will result in [0,1,2,2,3,4]

    Keep this in mind especially where complex types like
    dicts nest inside a list - a merge will always add a new dict list item,
    not merge it into whatever dicts might exist on the list already.

    For example, say input context is:
        key1: value1
        key2: value2
        key3:
            k31: value31
            k32: value32
        defaults:
            key2: 'aaa_{key1}_zzz'
            key3:
                k33: value33
            key4: 'bbb_{key2}_yyy'

    This will result in return context:
        key1: value1
        key2: value2
        key3:
            k31: value31
            k32: value32
            k33: value33
        key4: bbb_value2_yyy
    """
    logger.debug("started")
    context.assert_key_has_value(key='defaults', caller=__name__)

    context.set_defaults(context['defaults'])

    logger.info("set %s context item defaults.", len(context['defaults']))

    logger.debug("done")



"""Step that echos simple input back to logger."""
import logging

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Echo input to output. Input context['echoMe'].

    Args:
        context: dictionary-like. context is mandatory.
                 context must contain key 'echoMe'
                 context['echoMe'] will echo the value to logger.
                 This logger could well be stdout.

    When you execute the pipeline, it should look something like this:
    pypyr [name here] 'echoMe=test', assuming a keyvaluepair context parser.
    """
    logger.debug("started")

    assert context, ("context must be set for echo. Did you set "
                     "'echoMe=text here'?")

    context.assert_key_exists('echoMe', __name__)

    if isinstance(context['echoMe'], str):
        val = context.get_formatted('echoMe')
    else:
        val = context['echoMe']

    logger.notify(val)

    logger.debug("done")



"""pypyr step that parses file, does string replacement and writes output."""
import logging
from pypyr.steps.dsl.fileinoutrewriter import StreamReplacePairsRewriterStep

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Parse input file and replace a search string.

    This also does string substitutions from context on the fileReplacePairs.
    It does this before it search & replaces the in file.

    Be careful of order. If fileReplacePairs is not an ordered collection,
    replacements could evaluate in any given order. If this is coming in from
    pipeline yaml it will be an ordered dictionary, so life is good.

    Args:
        context: pypyr.context.Context. Mandatory.
                 The following context keys expected:
                - fileReplace
                    - in. mandatory.
                      str, path-like, or an iterable (list/tuple) of
                      strings/paths. Each str/path can be a glob, relative or
                      absolute path.
                    - out. optional. path-like.
                      Can refer to a file or a directory.
                      will create directory structure if it doesn't exist. If
                      in-path refers to >1 file (e.g it's a glob or list), out
                      path can only be a directory - it doesn't make sense to
                      write >1 file to the same single file (this is not an
                      appender.) To ensure out_path is read as a directory and
                      not a file, be sure to have the path separator (/) at the
                      end.
                      If out_path is not specified or None, will in-place edit
                      and overwrite the in-files.
                    - replacePairs. mandatory. Dictionary where items are:
                      'find_string': 'replace_string'

    Returns:
        None.

    Raises:
        FileNotFoundError: take a guess
        pypyr.errors.KeyNotInContextError: Any of the required keys missing in
                                          context.
        pypyr.errors.KeyInContextHasNoValueError: Any of the required keys
                                                  exists but is None.

    """
    logger.debug("started")

    StreamReplacePairsRewriterStep(__name__, 'fileReplace', context).run_step()

    logger.debug("done")



"""Continous Deployment logging functions.

Configuration for the python logging library.
"""
import logging


NOTIFY = 25


def set_logging_config(log_level, handlers):
    """Set python logging library config.

    Run this ONCE at the start of your process. It formats the python logging
    module's output.
    """
    if log_level is None:
        level = 25
        format_string = '%(message)s'
    else:
        level = log_level
        format_string = (
            '%(asctime)s %(levelname)s:%(name)s:%(funcName)s: %(message)s')

    logging.basicConfig(
        format=format_string,
        datefmt='%Y-%m-%d %H:%M:%S',
        level=level,
        handlers=handlers)


def notify(self, msg, *args, **kwargs):
    """Log a message with severity 'NOTIFY' on the root logger."""
    if self.isEnabledFor(NOTIFY):
        self._log(NOTIFY, msg, args, **kwargs)


def set_up_notify_log_level():
    """Add up a global notify severity to the python logging package.

    NOTIFY severity is logging level between INFO and WARNING.
    By default it outputs only echo step and step name
    with description.
    """
    # could (should?) be checking hasattr like so:
    # hasattr(logging, levelName):
    # hasattr(logging, methodName):
    # hasattr(logging.getLoggerClass(), methodName):
    # but this extra check is arguably *more* overhead than just assigning it?
    logging.addLevelName(NOTIFY, "NOTIFY")
    logging.NOTIFY = NOTIFY
    logging.getLoggerClass().notify = notify


def set_root_logger(log_level, log_path=None):
    """Set the root logger 'pypyr'. Do this before you do anything else.

    Run once and only once at initialization.
    """
    handlers = []

    console_handler = logging.StreamHandler()
    handlers.append(console_handler)
    if log_path:
        file_handler = logging.FileHandler(log_path)
        handlers.append(file_handler)

    set_logging_config(log_level, handlers=handlers)

    root_logger = logging.getLogger("pypyr")
    root_logger.debug(
        "Root logger %s configured with level %s",
        root_logger.name, log_level)



"""Version information."""

import platform

__version__ = '4.2.0'


def get_version():
    """Return package-name __version__ python python_version."""
    return (f'pypyr {__version__} '
            f'python {platform.python_version()}')



"""pypyr step yaml definition for control of flow instructions."""
import logging
from pypyr.errors import (ContextError,
                          KeyInContextHasNoValueError,
                          KeyNotInContextError)


def control_of_flow_instruction(name, instruction_type, context, context_key):
    """Run a control of flow instruction.

    The step config in the context dict looks like this:
        <<instruction-name>>: <<cmd string>>. Mandatory.

        OR, as a dict
        <<instruction-name:
            groups: <<str>> or <<list of str>> - mandatory.
            success: <<str>>
            failure: <<str>>

    Args:
        name: Unique name for step. Likely __name__ of calling step.
        instruction_type: Type - must inherit from
                          pypyr.errors.ControlOfFlowInstruction
        context: pypyr.context.Context. Look for config in this context
                 instance.
        context_key: str name of step config in context.

    """
    assert name, ("name parameter must exist for a ControlOfFlowStep.")
    assert context, ("context param must exist for ControlOfFlowStep.")
    # this way, logs output as the calling step, which makes more sense
    # to end-user than a mystery steps.dsl.blah logging output.
    logger = logging.getLogger(name)
    logger.debug("starting")

    context.assert_key_has_value(key=context_key, caller=name)
    original_config = (context_key, context[context_key])

    config = context.get_formatted(context_key)

    if isinstance(config, str):
        groups = [config]
        success_group = None
        failure_group = None
    elif isinstance(config, list):
        groups = config
        success_group = None
        failure_group = None
    elif isinstance(config, dict):
        if 'groups' not in config:
            raise KeyNotInContextError(
                f"{context_key} needs a child key 'groups', which should "
                "be a list or a str with the step-group name(s) you want "
                f"to run. This is for step {name}.")
        groups = config['groups']
        if not groups:
            raise KeyInContextHasNoValueError(
                f"{context_key}.groups must have a value for step {name}")

        if isinstance(groups, str):
            groups = [groups]

        success_group = config.get('success', None)
        failure_group = config.get('failure', None)
    else:
        raise ContextError(
            f"{context_key} needs a child key 'groups', which should "
            "be a list or a str with the step-group name(s) you want "
            f"to run. This is for step {name}. Instead, you've got {config}")

    if success_group is not None and not isinstance(success_group, str):
        raise ContextError(
            f"{context_key}.success must be a string for {name}.")

    if failure_group is not None and not isinstance(failure_group, str):
        raise ContextError(
            f"{context_key}.failure must be a string for {name}.")

    logger.info(
        ("step %s about to hand over control with %s: Will run groups: %s "
         " with success %s and failure %s"),
        name,
        context_key,
        groups,
        success_group,
        failure_group)
    raise instruction_type(groups=groups,
                           success_group=success_group,
                           failure_group=failure_group,
                           original_config=original_config)



"""pypyr caching base class and functions."""
import logging
import threading

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)


class Cache():
    """Thread-safe general purpose cache for objects.

    Add things to the cache by calling get(key, creator). If the requested key
    doesn't exist, will add the item to the cache for you.
    """

    def __init__(self):
        """Instantiate the cache."""
        self._lock = threading.RLock()
        self._cache = {}

    def clear(self):
        """Clear the cache of all objects."""
        with self._lock:
            self._cache.clear()

    def get(self, key, creator):
        """Get key from cache. If key not exist, call creator and cache result.

        Looks for key in cache and returns object for that key.

        If key is not found, call creator and save the result to cache for that
        key.

        Be warned that get happens under the context of a RLock. . . so if
        creator takes a long time you might well be blocking.

        Args:
            key: key (unique id) of cached item
            creator: callable that will create cached object if key not found

        Returns:
            Cached item at key or the result of creator()
        """
        with self._lock:
            if key in self._cache:
                logger.debug("%s loading from cache", key)
                obj = self._cache[key]
            else:
                logger.debug("%s not found in cache. . . creating", key)
                obj = creator()
                self._cache[key] = obj

        return obj



"""Default execution entry point if running the package via python -m."""
import pypyr.cli
import sys


def main():
    """Run pypyr from script entry point."""
    return pypyr.cli.main()


if __name__ == '__main__':
    sys.exit(main())



"""cli entry point for pipeline runner.

Parse command line arguments in, invoke pipelinerunner.
"""
import argparse
from pathlib import Path
import pypyr.log.logger
import pypyr.pipelinerunner
import pypyr.version
import signal
import sys
import textwrap
import traceback


def wrap(text, **kwargs):
    """Wrap lines in argparse so they align nicely in 2 columns.

    Default width is 70.

    With gratitude to paul.j3 https://bugs.python.org/issue12806
    """
    # apply textwrap to each line individually
    text = text.splitlines()
    text = [textwrap.fill(line, **kwargs) for line in text]
    return '\n'.join(text)


def get_args(args):
    """Parse arguments passed in from shell."""
    return get_parser().parse_args(args)


def get_parser():
    """Return ArgumentParser for pypyr cli."""
    parser = argparse.ArgumentParser(
        allow_abbrev=True,
        description='pypyr pipeline runner',
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('pipeline_name',
                        help=wrap('Name of pipeline to run. It should exist '
                                  'in the ./pipelines directory.'))
    parser.add_argument(dest='context_args',
                        nargs='*',
                        default=None,
                        help=wrap('Initialize context with this. Parsed by '
                                  'the pipeline\'s context_parser function.\n'
                                  'Separate multiple args with spaces.'))
    parser.add_argument('--groups', dest='groups',
                        nargs='*',
                        default=None,
                        help=wrap(
                            'Step-Groups to run. defaults to "steps".\n'
                            'You probably want to order --groups AFTER the '
                            'pipeline name and context positional args. e.g\n'
                            'pypyr pipename context --groups group1 group2\n'
                            'If you prefer putting them before, use a -- to '
                            'separate groups from the pipeline name, e.g\n'
                            'pypyr --groups group1 group2 -- pipename context')
                        )
    parser.add_argument('--success', dest='success_group', default=None,
                        help=wrap(
                            'Step-Group to run on successful completion of '
                            'pipeline.\n'
                            'Defaults to "on_success"'))
    parser.add_argument('--failure', dest='failure_group', default=None,
                        help=wrap(
                            'Step-Group to run on error completion of '
                            'pipeline.\n'
                            'Defaults to "on_failure"'))
    parser.add_argument('--dir', dest='working_dir', default=Path.cwd(),
                        help=wrap('Working directory. Use if your pipelines '
                                  'directory is elsewhere.\n'
                                  'Defaults to cwd.'))
    parser.add_argument('--log', '--loglevel', dest='log_level', type=int,
                        default=None,
                        help=wrap(
                            'Integer log level. Defaults to 25 (NOTIFY).\n'
                            '10=DEBUG \n'
                            '20=INFO\n'
                            '25=NOTIFY\n'
                            '30=WARNING\n'
                            '40=ERROR\n'
                            '50=CRITICAL\n'
                            'Log Level < 10 gives full traceback on errors.'))
    parser.add_argument('--logpath', dest='log_path',
                        help=wrap(
                            'Log-file path. Append log output to this path.'))
    parser.add_argument('--version', action='version',
                        help='Echo version number.',
                        version=f'{pypyr.version.get_version()}')
    return parser


def main(args=None):
    """Entry point for pypyr cli.

    The setup_py entry_point wraps this in sys.exit already so this effectively
    becomes sys.exit(main()).
    The __main__ entry point similarly wraps sys.exit().
    """
    if args is None:
        args = sys.argv[1:]

    parsed_args = get_args(args)

    try:
        pypyr.log.logger.set_root_logger(log_level=parsed_args.log_level,
                                         log_path=parsed_args.log_path)

        return pypyr.pipelinerunner.main(
            pipeline_name=parsed_args.pipeline_name,
            pipeline_context_input=parsed_args.context_args,
            working_dir=parsed_args.working_dir,
            groups=parsed_args.groups,
            success_group=parsed_args.success_group,
            failure_group=parsed_args.failure_group)
    except KeyboardInterrupt:
        # Shell standard is 128 + signum = 130 (SIGINT = 2)
        sys.stdout.write("\n")
        return 128 + signal.SIGINT
    except Exception as e:
        # stderr and exit code 255
        sys.stderr.write("\n")
        sys.stderr.write(f"\033[91m{type(e).__name__}: {str(e)}\033[0;0m")
        sys.stderr.write("\n")
        # at this point, you're guaranteed to have args and thus log_level
        if parsed_args.log_level:
            if parsed_args.log_level < 10:
                # traceback prints to stderr by default
                traceback.print_exc()

        return 255



"""Load pipelines from local disk."""
import logging
from pathlib import Path
from pypyr.errors import PipelineNotFoundError
import pypyr.moduleloader
import pypyr.yaml

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)


def get_pipeline_path(pipeline_name, working_directory):
    """Look for the pipeline in the various places it could be.

    First checks the cwd
    Then check cwd/pipelines
    Then checks {pypyr install dir}/pipelines dir.

    Args:
        pipeline_name: string. Name of pipeline to find
        working_directory: string. Path in which to look for pipeline_name.yaml

    Returns:
        Absolute path to the pipeline_name.yaml file

    Raises:
        PipelineNotFoundError: if pipeline_name.yaml not found in working_dir
                               or in {pypyr install dir}/pipelines.

    """
    logger.debug("starting")

    # look for name.yaml in the pipelines/ sub-directory
    logger.debug("current directory is %s", working_directory)

    cwd = Path(working_directory)

    # look for cwd/{pipeline_name}.yaml
    pipeline_path = cwd.joinpath(f'{pipeline_name}.yaml')

    if pipeline_path.is_file():
        logger.debug("Found %s", pipeline_path)
    else:
        logger.debug("%s not found in working directory. "
                     "Looking in '{working dir}/pipelines' instead.",
                     pipeline_name)
        # looking for {cwd}/pipelines/[pipeline_name].yaml
        pipeline_path = cwd.joinpath('pipelines',
                                     f'{pipeline_name}.yaml').resolve()

        if pipeline_path.is_file():
            logger.debug("Found %s", pipeline_path)
        else:
            logger.debug("%s not found in working directory/pipelines folder. "
                         "Looking in pypyr install directory instead.",
                         pipeline_name)
            pypyr_dir = Path(__file__).resolve().parents[1]
            logger.debug("pypyr installation directory is: %s", pypyr_dir)
            pipeline_path = pypyr_dir.joinpath('pipelines',
                                               f'{pipeline_name}.yaml')

            if pipeline_path.is_file():
                logger.debug("Found %s", pipeline_path)
            else:
                raise PipelineNotFoundError(
                    f"{pipeline_name}.yaml not found in any of the "
                    "following:\n"
                    f"{working_directory}\n"
                    f"{working_directory}/pipelines\n"
                    f"{pypyr_dir}/pipelines")

    logger.debug("done")
    return pipeline_path


def get_pipeline_definition(pipeline_name, working_dir):
    """Open and parse the pipeline definition yaml.

    Parses pipeline yaml and returns dictionary representing the pipeline.

    pipeline_name.yaml should be in the working_dir/ directory, or in the
    fileloader directory look-up sequence.

    Args:
        pipeline_name: string. Name of pipeline. This will be the file-name of
                       the pipeline - i.e {pipeline_name}.yaml
        working_dir: path. Start looking in
                           ./working_dir/pipeline_name.yaml

    Returns:
        dict describing the pipeline, parsed from the pipeline yaml.

    Raises:
        FileNotFoundError: pipeline_name.yaml not found in the various pipeline
                           dirs.

    """
    logger.debug("starting")

    pipeline_path = get_pipeline_path(
        pipeline_name=pipeline_name,
        working_directory=working_dir)

    logger.debug("Trying to open pipeline at path %s", pipeline_path)
    try:
        with open(pipeline_path) as yaml_file:
            pipeline_definition = pypyr.yaml.get_pipeline_yaml(
                yaml_file)
            logger.debug(
                "found %d stages in pipeline.", len(pipeline_definition))
    except FileNotFoundError:
        logger.error(
            "The pipeline doesn't exist. Looking for a file here: "
            "%s", pipeline_path)
        raise

    logger.debug("pipeline definition loaded")

    logger.debug("done")
    return pipeline_definition



"""Utility methods for test."""
import logging
from contextlib import contextmanager
from copy import deepcopy
from unittest.mock import MagicMock, Mock


class DeepCopyMagicMock(MagicMock):
    """Derive a new MagicMock doing a deepcopy of args to calls.

    MagicMocks store a reference to a mutable object - so on multiple calls to
    the mock the call history isn't maintained if the same obj mutates as an
    arg to those calls. https://bugs.python.org/issue33667

    It's probably not sensible to deepcopy all mock calls. So this little class
    is for patching the MagicMock class specifically, where it will do the
    deepcopy only where specifically patched.

    See here:
    https://docs.python.org/3/library/unittest.mock-examples.html#coping-with-mutable-arguments
    """

    def __call__(self, *args, **kwargs):
        """Mock call override does deepcopy."""
        return super(DeepCopyMagicMock, self).__call__(*deepcopy(args),
                                                       **deepcopy(kwargs))


@contextmanager
def patch_logger(name, level=logging.DEBUG):
    """Add a new mocked handler to the logger.

    Will store formatted log records with the only one specified level.

    If logging level is DEBUG, will take care only about DEBUG
    records and will not report other levels even higher(INFO, ERROR and etc).
    """
    logger = logging.getLogger(name)
    logging_level = logger.level
    if level < logging_level or not logging_level:
        logger.setLevel(level)

    mock = Mock()

    class MockHandler(logging.NullHandler):
        def handle(self, record):
            if record.levelno == level:
                mock(self.format(record))

    handler = MockHandler(level)
    logger.addHandler(handler)

    try:
        yield mock
    finally:
        # restore logger settings
        logger.removeHandler(handler)
        logger.setLevel(logging_level)


def read_file_to_list(path):
    """Read path and write each line to list.

    Args:
        path: path-like. Path to file to read.

    Returns:
        list of str. Each line in file is an item on the list.
    """
    with open(path, 'r') as file:
        return [line.rstrip() for line in file.readlines()]



"""yaml handling functions."""
import ruamel.yaml as yamler
from pypyr.context import Context
from pypyr.dsl import Jsonify, PyString, SicString


def get_pipeline_yaml(file):
    """Return pipeline yaml from open file object.

    Use specific custom representers to model the custom pypyr pipeline yaml
    format, to load in special literal types like py and sic strings.

    If looking to extend the pypyr pipeline syntax with special types, add
    these to the tag_representers list.

    Args:
        file: open file-like object.

    Returns:
        dict-like representation of loaded yaml.

    """
    tag_representers = [Jsonify, PyString, SicString]

    yaml_loader = get_yaml_parser_roundtrip()

    for representer in tag_representers:
        yaml_loader.register_class(representer)

    pipeline_definition = yaml_loader.load(file)
    return pipeline_definition


def get_yaml_parser_safe():
    """Create the safe yaml parser object with this factory method.

    The safe yaml parser does NOT resolve unknown tags.

    Returns:
        ruamel.yaml.YAML object with safe loader

    """
    return yamler.YAML(typ='safe', pure=True)


def get_yaml_parser_roundtrip():
    """Create the yaml parser object with this factory method.

    The round-trip parser preserves:
    - comments
    - block style and key ordering are kept, so you can diff the round-tripped
      source
    - flow style sequences ( ‘a: b, c, d’) (based on request and test by
      Anthony Sottile)
    - anchor names that are hand-crafted (i.e. not of the form``idNNN``)
    - merges in dictionaries are preserved

    Returns:
        ruamel.yaml.YAML object with round-trip loader

    """
    yaml_writer = yamler.YAML(typ='rt', pure=True)
    # if this isn't here the yaml doesn't format nicely indented for humans
    yaml_writer.indent(mapping=2, sequence=4, offset=2)
    return yaml_writer


def get_yaml_parser_roundtrip_for_context():
    """Create a yaml parser that can serialize the pypyr Context.

    Create yaml parser with get_yaml_parser_roundtrip, adding Context.
    This allows the yaml parser to serialize the pypyr Context.
    """
    yaml_writer = get_yaml_parser_roundtrip()

    # Context is a dict data structure, so can just use a dict representer
    yaml_writer.Representer.add_representer(
        Context,
        yamler.representer.RoundTripRepresenter.represent_dict)

    return yaml_writer



"""Global cache for pipeline yaml.

Attributes:
    pipeline_cache: global instance of the pipeline yaml cache.
                    Use this attribute to access the cache from elsewhere.
"""
import logging
from pypyr.cache.cache import Cache
from pypyr.cache.loadercache import pypeloader_cache
import pypyr.moduleloader

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)


class PipelineCache(Cache):
    """Get and add Pipeline yaml from the pipelines cache."""

    def get_pipeline(self, pipeline_name, loader=None):
        """Get cached pipeline yaml. Adds to cache if not exist.

        Args:
            pipeline_name: (string) Name of pipeline, sans .yaml at end.
            loader: (string) Name of loader to load the pipeline

        Returns:
            yaml: Yaml representation of pipeline_name
        """
        logger.debug("starting")
        pipeline = self.get(pipeline_name, load_pipeline(pipeline_name,
                                                         loader))

        logger.debug("done")
        return pipeline


# single global instance of pipelines in a cache
pipeline_cache = PipelineCache()


def load_pipeline(pipeline_name, loader=None):
    """Return function that loads the pipeline using loader.

    This is a wrapped function, so it returns a callable. The actual work
    happens in the inner function.

    Args:
        pipeline_name (str): Name of pipeline, sans .yaml at end.
        loader (str): str. optional. Absolute name of pipeline loader module.
                If not specified will use pypyr.pypeloaders.fileloader.

    Returns:
        callable that will execute the loader and retrieve the pipeline def.
    """
    def load_pipeline_inner():
        logger.debug("starting")

        get_pipeline_definition = pypeloader_cache.get_pype_loader(loader)

        pipeline_definition = get_pipeline_definition(
            pipeline_name=pipeline_name,
            working_dir=pypyr.moduleloader.get_working_directory()
        )
        logger.debug("done")
        return pipeline_definition

    return load_pipeline_inner



"""Context parser that returns a dictionary from a json string."""
from collections.abc import Mapping
import json
import logging

logger = logging.getLogger(__name__)


def get_parsed_context(args):
    """Parse input context args and returns context as dictionary."""
    logger.debug("starting")
    if not args:
        logger.debug("pipeline invoked without context arg set. For "
                     "this json parser you're looking for something "
                     "like: "
                     "pypyr pipelinename '{\"key1\":\"value1\","
                     "\"key2\":\"value2\"}'")
        return None

    # deserialize the input context string into json
    payload = json.loads(' '.join(args))

    if not isinstance(payload, Mapping):
        raise TypeError("json input should describe an object at the top "
                        "level. You should have something like \n"
                        "{\n\"key1\":\"value1\",\n\"key2\":\"value2\"\n}\n"
                        "at the json top-level, not an [array] or literal.")

    return payload



"""pypyr step that pretty prints debug information to console.

Dumps the pypyr context to stdout. This may assist in debugging when trying to
see what values are what.

All inputs are optional. This means you can run debug in a pipeline just with
- pypyr.steps.debug
In this case it will dump the entire context as is without applying formatting.

Debug supports the following optional inputs:
debug:
    keys: str for a single key name to dump. Or a list of key names to dump.
    format: Boolean, defaults False. Applies formatting expressions on output.

"""
import pprint
import logging

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Print debug info to console.

    context is a dictionary or dictionary-like.

    If you use pypyr.steps.debug as a simple step (i.e you do NOT specify the
    debug input context), it will just dump the entire context to stdout.

    Configure the debug step with the following optional context item:
        debug:
            keys: str (for single key) or list (of str keys). Only dump the
                  specified keys.
            format: bool. Defaults False. Applies formatting expressions on
                    dump.
    """
    logger.debug("started")

    debug = context.get('debug', None)

    if debug:
        keys = debug.get('keys', None)
        format = debug.get('format', False)

        if keys:
            logger.debug("Writing to output: %s", keys)
            if isinstance(keys, str):
                payload = {keys: context[keys]}
            else:
                payload = {k: context[k] for k in keys}
        else:
            logger.debug(
                "No keys specified. Writing entire context to output.")
            payload = context

        if format:
            payload = context.get_formatted_value(payload)
    else:
        payload = context

    if logger.isEnabledFor(logging.INFO):
        # call pformat only if logging is enabled
        logger.info('\n%s', pprint.pformat(payload))

    logger.debug("done")



"""Step that stops current step-group."""
import logging
from pypyr.errors import StopStepGroup

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Stop current step-group."""
    logger.debug("started")

    logger.info("StopStepGroup: stopping step-group...")
    raise StopStepGroup("pypyr.steps.stopstepgroup stopped current step-group")



"""Utility functions for file system operations. Read, format files, write."""
from abc import ABC, abstractmethod
import glob
from itertools import chain
import json
import logging
import os
from pathlib import Path
from tempfile import NamedTemporaryFile
from pypyr.errors import Error
import pypyr.yaml

# pypyr logger means the log level will be set correctly and output formatted.
logger = logging.getLogger(__name__)


class FileRewriter(ABC):
    """FileRewriter reads input file, formats it and write to output file.

    Use this abstract base class to implement rewriters.

    This base class contains useful functionality to loop through input and
    output paths, leaving the handling of individual files up to the deriving
    classes.
    """

    def __init__(self, formatter):
        """Initialize formatter.

        Args:
            formatter: Callable object that will format the IN file payload to
                       create OUT file.
        """
        self.formatter = formatter

    @abstractmethod
    def in_to_out(self, in_path, out_path):
        """Take in_path, applies formatting, writes to out_path.

        Input arguments can be str or path-like. Relative or absolute paths
        will work.

        Args:
            in_path: str or path-like. Must refer to a single existing file.
            out_path: str or path-like. Must refer to a single destination file
                      location. will create directory structure if it doesn't
                      exist.
        Returns:
            None.

        """
        raise NotImplementedError(
            'you must implement in_to_out(in_path, out_path) for a '
            'FileFormatter')

    def files_in_to_out(self, in_path, out_path=None):
        """Write in files to out, calling the line_handler on each line.

        Calls file_in_to_out under the hood to format the in_path payload. The
        formatting processing is done by the self.formatter instance.

        Args:
            in_path: str, path-like, or an iterable (list/tuple) of
                     strings/paths. Each str/path can be a glob, relative or
                     absolute path.
            out_path: str or path-like. Can refer to a file or a directory.
                      will create directory structure if it doesn't exist. If
                      in-path refers to >1 file (e.g it's a glob or list), out
                      path can only be a directory - it doesn't make sense to
                      write >1 file to the same single file (this is no an
                      appender.) To ensure out_path is read as a directory and
                      not a file, be sure to have the path separator (/) at the
                      end.
                      Top tip: Path-like objects strip the trailing slash. If
                      you want to pass in a dir that does not exist yet as
                      out-path with a trailing /, you should be passing it as a
                      str to preserve the /.
                      If out_path is not specified or None, will in-place edit
                      and overwrite the in-files.

        Returns:
            None.

        """
        in_paths = get_glob(in_path)

        in_count = len(in_paths)
        if in_count == 0:
            logger.debug(f'in path found {in_count} paths.')
        else:
            logger.debug(f'in path found {in_count} paths:')
            for path in in_paths:
                logger.debug(f'{path}')
            logger.debug(
                'herewith ends the paths. will now process each file.')

        if in_paths:
            # derive the destination directory, ensure it's ready for writing
            basedir_out = None
            is_outfile_name_known = False
            if out_path:
                # outpath could be a file, or a dir
                pathlib_out = Path(out_path)
                # yep, Path() strips trailing /, hence check original string
                if isinstance(out_path, str) and out_path.endswith(os.sep):
                    # ensure dir - mimic posix mkdir -p
                    pathlib_out.mkdir(parents=True, exist_ok=True)
                    basedir_out = pathlib_out
                elif pathlib_out.is_dir():
                    basedir_out = pathlib_out
                else:
                    if len(in_paths) > 1:
                        raise Error(
                            f'{in_path} resolves to {len(in_paths)} files, '
                            'but you specified only a single file as out '
                            f'{out_path}. If the outpath is meant to be a '
                            'directory, put a / at the end.')

                    # at this point it must be a file (not dir) path
                    # make sure that the parent dir exists
                    basedir_out = pathlib_out.parent
                    basedir_out.parent.mkdir(parents=True, exist_ok=True)
                    is_outfile_name_known = True

            # loop through all the in files and write them to the out dir
            file_counter = 0
            is_edit = False
            for path in in_paths:
                actual_in = Path(path)
                # recursive glob returns dirs too, only interested in files
                if actual_in.is_file():
                    if basedir_out:
                        if is_outfile_name_known:
                            actual_out = pathlib_out
                        else:
                            # default to original src file name if only out dir
                            # specified without an out file name
                            actual_out = basedir_out.joinpath(actual_in.name)

                        logger.debug("writing %s to %s", path, actual_out)
                        self.in_to_out(in_path=actual_in, out_path=actual_out)
                    else:
                        logger.debug("editing %s", path)
                        self.in_to_out(in_path=actual_in)
                        is_edit = True
                    file_counter += 1

            if is_edit:
                logger.info(
                    "edited & wrote %s file(s) at %s", file_counter, in_path)
            else:
                logger.info(
                    "read %s, formatted and wrote %s file(s) to %s",
                    in_path, file_counter, out_path)
        else:
            logger.info("%s found no files", in_path)


class ObjectRewriter(FileRewriter):
    """Load a single file into an object, run formatter on it and write out.

    Object instantiation takes a formatter.
    writer = StreamRewriter(formatter)

    Formatter signature: iterable = formatter(iterable)
    It returns an iterator. The single input argument is an iterable.
    Tip, use function or callable object with __call__

    Object instantion also takes an ObjectRepresenter. An ObjectRepresenter
    has a load and a dump method that handles the object deserialization and
    serialization.
    """

    def __init__(self, formatter, object_representer):
        """Initialize formatter and object representer.

        Args:
            formatter: Callable object/function that will format object loaded
                       from in file. Formatter signature:
                       iterable = formatter(iterable)
            object_representer: An ObjectRepresenter instance.

        """
        super().__init__(formatter)
        self.object_representer = object_representer
        logger.debug('obj loader set')

    def in_to_out(self, in_path, out_path=None):
        """Load file into object, formats, writes object to out.

        If in_path and out_path point to the same thing it will in-place edit
        and overwrite the in path. Even easier, if you do want to edit a file
        in place, don't specify out_path, or set it to None.

        Args:
            in_path: str or path-like. Must refer to a single existing file.
            out_path: str or path-like. Must refer to a single destination file
                      location. will create directory structure if it doesn't
                      exist.
                      If out_path is not specified or None, will in-place edit
                      and overwrite the in-files.

        Returns:
            None.

        """
        if is_same_file(in_path, out_path):
            logger.debug(
                "in path and out path are the same file. writing to temp "
                "file and then replacing in path with the temp file.")
            out_path = None

        logger.debug("opening source file: %s", in_path)
        with open(in_path) as infile:
            obj = self.object_representer.load(infile)

        if out_path:
            logger.debug(
                f"opening destination file for writing: {out_path}")
            ensure_dir(out_path)
            with open(out_path, 'w') as outfile:
                self.object_representer.dump(outfile, self.formatter(obj))
            return
        else:
            logger.debug("opening temp file for writing...")
            with NamedTemporaryFile(mode='w+t',
                                    dir=os.path.dirname(in_path),
                                    delete=False) as outfile:
                self.object_representer.dump(outfile, self.formatter(obj))

            logger.debug("moving temp file to: %s", in_path)

            move_temp_file(outfile.name, infile.name)


class StreamRewriter(FileRewriter):
    """Streaming style in-to-out reader and writer.

    Reads IN file  line-by-line, formats each line and writes to OUT in a
    stream. You can expect memory use to stay more or less flat, depending on
    how big your lines are.

    Object instantiation takes a formatter.
    writer = StreamRewriter(formatter)

    Formatter signature: iterator = formatter(iterable)
    It returns an iterator. The single input argument is an iterable.
    Tip, use function or callable object with __call__

    """

    def in_to_out(self, in_path, out_path=None):
        """Write a single file in to out, running self.formatter on each line.

        If in_path and out_path point to the same thing it will in-place edit
        and overwrite the in path. Even easier, if you do want to edit a file
        in place, don't specify out_path, or set it to None.

        Args:
            in_path: str or path-like. Must refer to a single existing file.
            out_path: str or path-like. Must refer to a single destination file
                      location. will create directory structure if it doesn't
                      exist.
                      If out_path is not specified or None, will in-place edit
                      and overwrite the in-files.

        Returns:
            None.

        """
        is_in_place_edit = False
        if is_same_file(in_path, out_path):
            logger.debug(
                "in path and out path are the same file. writing to temp "
                "file and then replacing in path with the temp file.")
            out_path = None
            is_in_place_edit = True

        logger.debug("opening source file: %s", in_path)
        with open(in_path) as infile:
            if out_path:
                logger.debug(
                    "opening destination file for writing: %s", out_path)
                ensure_dir(out_path)
                with open(out_path, 'w') as outfile:
                    outfile.writelines(self.formatter(infile))
                return
            else:
                logger.debug("opening temp file for writing...")
                with NamedTemporaryFile(mode='w+t',
                                        dir=os.path.dirname(in_path),
                                        delete=False) as outfile:
                    outfile.writelines(self.formatter(infile))

                is_in_place_edit = True

        # only replace infile AFTER it's closed, outside the with.
        # pragma exclude because func actually returns on 287 in if out_path,
        # and cov not smart enough to realize that !is_in_place_edit won't ever
        # happen here (the function will have exited already)
        if is_in_place_edit:    # pragma: no branch
            logger.debug("moving temp file to: %s", in_path)
            move_temp_file(outfile.name, infile.name)


class ObjectRepresenter(ABC):
    """Abstract base class to handle object serialization and deserialization.

    Derive from this base class to create your own object representers that can
    serialize and deserialize a payload.

    This tends to be useful for the ObjectRewriter.
    """

    @abstractmethod
    def load(self, file):
        """Instantiate object from input file."""
        pass

    @abstractmethod
    def dump(self, file, payload):
        """Serialize payload to file."""
        pass


class JsonRepresenter(ObjectRepresenter):
    """Load and dump a json payload. Useful for ObjectRewriter."""

    def load(self, file):
        """Load Json object from open file input.

        Args:
            file: Open file-like object.

        Returns:
            None.

        """
        return json.load(file)

    def dump(self, file, payload):
        """Dump json oject to open file output.

        Writes json with 2 spaces indentation.

        Args:
            file: Open file-like object. Must be open for writing.
            payload: The Json object to write to file.

        Returns:
            None.

        """
        json.dump(payload, file, indent=2, ensure_ascii=False)


class YamlRepresenter(ObjectRepresenter):
    """Load and dump a Yaml payload. Useful for ObjectRewriter."""

    def __init__(self):
        """Instantiate the yaml representer."""
        self.yaml_parser = pypyr.yaml.get_yaml_parser_roundtrip()

    def load(self, file):
        """Load Yaml object from open file input.

        This uses a safe loader, so only recognized yaml types allowed.

        Args:
            file: Open file-like object.

        Returns:
            None.

        """
        return self.yaml_parser.load(file)

    def dump(self, file, payload):
        """Dump yaml oject to open file output.

        Writes yaml with 2 spaces indentation.

        Args:
            file: Open file-like object. Must be open for writing.
            payload: The yaml object to write to file.

        Returns:
            None.

        """
        self.yaml_parser.dump(payload, file)


def ensure_dir(path):
    """Create all parent directories of path if they don't exist.

    Args:
        path. Path-like object. Create parent dirs to this path.

    Return:
        None.

    """
    os.makedirs(os.path.abspath(os.path.dirname(path)), exist_ok=True)


def get_glob(path):
    """Process the input path, applying globbing and formatting.

    Do note that this will returns files AND directories that match the glob.

    No tilde expansion is done, but *, ?, and character ranges expressed with
    [] will be correctly matched.

    Escape all special characters ('?', '*' and '['). For a literal match, wrap
    the meta-characters in brackets. For example, '[?]' matches the character
    '?'.

    If passing in an iterable of paths, will expand matches for each path in
    the iterable. The function will return all the matches for each path
    glob expression combined into a single list.

    Args:
        path: Path-like string, or iterable (list or tuple ) of paths.

    Returns:
        Combined list of paths found for input glob.

    """
    if isinstance(path, str):
        return glob.glob(path, recursive=True)
    if isinstance(path, os.PathLike):
        # hilariously enough, glob doesn't like path-like. Gotta be str.
        return glob.glob(str(path), recursive=True)
    elif isinstance(path, (list, tuple)):
        # each glob returns a list, so chain all the lists into one big list
        return list(chain.from_iterable(
            glob.glob(str(p), recursive=True) for p in path))
    else:
        raise TypeError("path should be string, path-like or a list. Instead, "
                        f"it's a {type(path)}")


def is_same_file(path1, path2):
    """Return True if path1 is the same file as path2.

    The reason for this dance is that samefile throws if either file doesn't
    exist.

    Args:
        path1: str or path-like.
        path2: str or path-like.

    Returns:
        bool. True if the same file, False if not.

    """
    return (
        path1 and path2
        and os.path.isfile(path1) and os.path.isfile(path2)
        and os.path.samefile(path1, path2))


def move_file(src, dest):
    """Move source file to destination.

    Overwrites dest.

    Args:
        src: str or path-like. source file
        dest: str or path-like. destination file

    Returns:
        None.

    Raises:
        FileNotFoundError: out path parent doesn't exist.
        OSError: if any IO operations go wrong.

    """
    try:
        os.replace(src, dest)
    except Exception as ex_replace:
        logger.error("error moving file %s to "
                     "%s. %s", src, dest, ex_replace)
        raise


def move_temp_file(src, dest):
    """Move src to dest. Delete src if something goes wrong.

    Overwrites dest.

    Args:
        src: str or path-like. source file
        dest: str or path-like. destination file

    Returns:
        None.

    Raises:
        FileNotFoundError: out path parent doesn't exist.
        OSError: if any IO operations go wrong. Does its best to clean up after
                 itself and remove temp files.

    """
    try:
        move_file(src, dest)
    except Exception:
        try:
            os.remove(src)
        except Exception as ex_clean:
            # at this point, something's deeply wrong, so log error.
            # raising the original error, though, not this error in the
            # error handler, as the 1st was the initial cause of all of
            # this.
            logger.error("error removing temp file %s. %s",
                         src, ex_clean)

        raise



"""Context parser that returns a dictionary from a key-value pair string.

Takes multiple input arguments, where each args is a key=value pair string and
returns a dictionary where each pair becomes a dictionary element, as children
of a dict key argDict.

Don't have spaces in your values unless you really meant it. "k1=v1, ' k2'=v2"
will result in a context key name of ' k2' not 'k2'.

So a argument input string like this "pig=ham hen=eggs yummypig=bacon", will
yield:
{'argDict': {'pig': 'ham', 'hen': ''eggs', 'yummypig': 'bacon'}}
"""
import logging

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)


def get_parsed_context(args):
    """Parse input context args and returns context as dictionary."""
    logger.debug("starting")
    if not args:
        logger.debug("pipeline invoked without context arg set. For this "
                     "dict parser you can use something "
                     "like:\n"
                     "pypyr pipelinename key1=value1 key2=value2")
        return {'argDict': None}

    # for each input argument, project key=value
    return {'argDict':
            dict(element.split('=') for element in args)}



"""pypyr step that sets context values from already existing context values.

This is handy if you need to prepare certain keys in context where a next step
might need a specific key. If you already have the value in context, you can
create a new key (or update existing key) with that value.

So let's say you already have context['currentKey'] = 'eggs'.
If you run newKey: currentKey, you'll end up with context['newKey'] == 'eggs'
"""
import logging

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Create new context keys from already existing context keys.

    Context is a dictionary or dictionary-like.
    context['contextSet'] must exist. It's a dictionary.
    Will iterate context['contextSet'] and save the values as new keys to the
    context.

    For example, say input context is:
        key1: value1
        key2: value2
        key3: value3
        contextSet:
            key2: key1
            key4: key3

    This will result in return context:
        key1: value1
        key2: value1
        key3: value3
        key4: value3
    """
    logger.debug("started")
    deprecated(context)
    context.assert_key_has_value(key='contextSet', caller=__name__)

    for k, v in context['contextSet'].items():
        logger.debug("setting context %s to value from context %s", k, v)
        context[k] = context[v]

    logger.debug("done")


def deprecated(context):
    """Handle deprecated context input."""
    logger.warning("contextSet is deprecated. It will "
                   "stop working upon the next major release. "
                   "Use the new pypyr.steps.contextcopy instead. It's a lot "
                   "better, promise!")



"""Substitution & interpolation formatting."""
from collections.abc import Mapping, Set, Sequence
from string import Formatter


class RecursionSpec():
    """Parse a string formatting spec.

    This class just parses a format_spec string. It looks for a leading 'ff' or
    'rf'.

    A format expression looks like this: {field_name!conversion:format_spec}

    The format_spec is everything after the ':'.

    It makes no attempt to implement business logic around what to do with the
    formatting string.

    Attributes:
        format_spec: The original format_spec without the leading 'ff' or 'rf'.
                     If is_set is False, is identical to input format_spec.
        has_recursed (bool): True if recursive operation completed for this
                             expression.
        is_flat (bool): True if format_spec starts with 'ff'.
        is_set (bool): True if format_spec is either 'ff' or 'rf'.
        is_recursive (bool): True if format_spec starts with 'rf'.
    """

    def __init__(self, format_spec):
        """Parse format_spec for recursion specifier.

        Args:
            format_spec (str): format specification parsed from string
                               formatting expression. Likely from Formatter's
                               .parse method.
        """
        recursion_spec = format_spec[:2]
        self.has_recursed = False
        self.is_set = False
        self.is_recursive = False
        self.is_flat = False

        if recursion_spec == 'rf':
            self.is_set = self.is_recursive = True
        elif recursion_spec == 'ff':
            self.is_set = self.is_flat = True

        self.format_spec = format_spec[2:] if self.is_set else format_spec


class RecursiveFormatter(Formatter):
    """Recursive or flat string formatting using standard python syntax.

    This maintains all the functionality of the standard python Formatter, only
    adding additional recursion to deal with nested formatting expressions. You
    can use conversions and format_spec as per usual. You can derive your own
    class and implement the standard overloads like get_field() and get_value()
    yourself.

    Full documentation here:
    https://docs.python.org/3/library/string.html#format-string-syntax

    A format expression looks like this: {field_name!conversion:format_spec}

    Both conversion and format_spec are optional.

    Recursive format means that if field_name resolves to a string that
    contains another formatting expression (a string with {} in it), it will
    keep formatting the result recursively until the result is a literal or
    object without any formatting expressions.

    Differs from the standard Python string.Formatter class in these ways:
    - If the input is an iterable object, iterates it recursively & applies
      formatting on all string types it finds in it.
    - Starting a format_spec with 'ff' formats once. (This is the default
      python string.Formatter behavior also).
    - Starting a format_spec with 'rf' formats recursively, until it
      encounters a `ff`.
    - Where a string has no format_spec, and contains a combination of
      literal and expression values, do standard flat format. For example,
      'arb {key} more arb' will format key once and cast the resulting value to
      string.
    - Where a string has no format_spec, and contains one and only one
      expression and nothing else, format recursively and keep the type of the
      source object. So where a formatting expression is '{key}', format
      recursively and return the ultimate value that key resolves to without
      casting it to string.

    You can still use all the usual format_spec functionality by adding the
    specifiers immediately after the 'ff' or 'rf'.

    Attributes:
        passthrough_types (tuple of type): Objects of this type do not format
                                           at all - pass through without
                                           applying any formatting logic.
        special_types (tuple of type): Objects of this type invoke
                                       .get_value(kwargs) on the object to get
                                       output value during formatting.
    """

    _FORMAT_SPEC_RECURSION_DEPTH = 2

    def __init__(self, passthrough_types=None, special_types=None):
        """Initialize me.

        Args:
            passthrough_types (tuple of type): Objects of this type do not
                                               format at all - pass through
                                               without applying any formatting
                                               logic. If single type, no need
                                               to put it in tuple.
            special_types (tuple of type): Objects of this type invoke
                                           .get_value on the instance to get
                                           output value during formatting. If
                                           single type, no need to put it in
                                           tuple.
        """
        self.passthrough_types = passthrough_types
        self.special_types = special_types

    def format(self, format_string, *args, **kwargs):
        """Format the input with arbitrary positional & keyword args.

        Use exactly as you would in string.Formatter. Only you get 'ff' & 'rf'
        extra.

        format_string can be any type, not just string. If format_string
        contains an iterable object, will iterate through it recursively and
        apply formatting on all strings as it goes.
        """
        return self.vformat(format_string, args, kwargs)

    def vformat(self, format_string, args, kwargs):
        """Format the input with predefined dict of args.

        Use exactly as you would in string.Formatter. Only you get 'ff' & 'rf'
        extra.

        Use me to avoid unpacking & repacking the dict as individual args
        using the *args & **kwargs syntax of .format().

        format_string can be any type, not just string. If format_string
        contains an iterable object, will iterate through it recursively and
        apply formatting on all strings as it goes.

        Args:
            format_string (object): Any object to format.
            args (iterable): Variable list positional args for formatting
                             expression.
            kwargs (iterable): Variable mapping of keyword args for formatting
                               expression.
        """
        used_args = set()
        result = self._get_formatted_iterable(format_string,
                                              args,
                                              kwargs,
                                              used_args)
        self.check_unused_args(used_args, args, kwargs)
        return result

    def _format_keep_type(self, format_string, args, kwargs, used_args,
                          recursion_depth, auto_arg_index=0,
                          is_recursive=False):
        """Do the actual implementation for formatting formattable objects.

        Don't call me directly. Use format() or vformat().

        The logic is largely poached from the python string.Formatter
        _vformat() method, with thanks!

        Args:
            format_string (object): Any object to format.
            args (iterable): Variable list positional args for formatting
                             expression.
            kwargs (iterable): Variable mapping of keyword args for formatting
                               expression.
            used_args (set): Mutating set of args used during formatting.
            recursion_depth (int): Only used if format_spec has formatting
                                   itself.
            auto_arg_index (int): Only used if format_spec has formatting
                                  itself.
            is_recursive (bool): Used internally to specify that
                                 in an active recursion loop where the parent
                                 indicated recursion. This is to allow nested
                                 formatting expressions to recurse themselves
                                 by default.
        """
        if recursion_depth < 0:
            raise ValueError('Max string recursion exceeded')
        result = []
        for literal_text, field_name, format_spec, conversion in \
                self.parse(format_string):

            # output the literal text
            if literal_text:
                result.append((literal_text, True, None))

            # if there's a field, output it
            if field_name is not None:
                # handle arg indexing when empty field_names are given.
                if field_name == '':
                    if auto_arg_index is False:
                        raise ValueError('cannot switch from manual field '
                                         'specification to automatic field '
                                         'numbering')
                    field_name = str(auto_arg_index)
                    auto_arg_index += 1
                elif field_name.isdigit():
                    if auto_arg_index:
                        raise ValueError('cannot switch from manual field '
                                         'specification to automatic field '
                                         'numbering')
                    # disable auto arg incrementing, if it gets
                    # used later on, then an exception will be raised
                    auto_arg_index = False

                # given the field_name, find the object it references
                # and the argument it came from
                obj, arg_used = self.get_field(field_name, args, kwargs)
                used_args.add(arg_used)

                # expand the format spec, if needed.
                # format spec expansion uses standard formatting of base class.
                format_spec, auto_arg_index = self._vformat(
                    format_spec, args, kwargs,
                    used_args, recursion_depth - 1,
                    auto_arg_index=auto_arg_index)

                # not doing this in format_field because need to know whether
                # to recurse or not here already - format_field doesn't take
                # the args/kwargs that get_formatted_iterable needs.
                recursion_spec = RecursionSpec(format_spec)

                # the resulting object could be formattable itself
                if recursion_spec.is_recursive or (
                        is_recursive and not recursion_spec.is_flat):
                    obj = self._get_formatted_iterable(
                        obj, args, kwargs, used_args, None, True)
                    recursion_spec.has_recursed = True

                # do any conversion on the resulting object
                obj = self.convert_field(obj, conversion)

                # only decide whether to format once sure that this is a
                # string and not a single object. thus, add to list, deal with
                # that after this field iteration completes.
                result.append((obj, False, recursion_spec))

        # where input is '{expr}' - i.e comprised entirely of 1 expression,
        # return the result object WITHOUT casting to string.
        if len(result) == 1:
            # single object. don't format for literals.
            obj, is_literal, recursion_spec = result[0]
            if is_literal:
                return obj
            else:
                if not (recursion_spec.has_recursed or recursion_spec.is_flat):
                    # default is go recursive on special case where there's a
                    # single formatting expression comprising the entire string
                    obj = self._get_formatted_iterable(
                        obj, args, kwargs, used_args, None,
                        recursion_spec.is_recursive)

                # if format_spec explicitly specified, can assume caller DOES
                # want the string conversion that format_spec does.
                if recursion_spec.format_spec:
                    return self.format_field(obj, recursion_spec.format_spec)
                else:
                    # do not format so retain the type
                    return obj

        else:
            # it must be a string with multiple formatting expressions in it
            return ''.join([obj
                            if is_literal
                            else self.format_field(obj,
                                                   recursion_spec.format_spec)
                            for obj, is_literal, recursion_spec in result])

    def _get_formatted_iterable(self, obj, args, kwargs, used_args, memo=None,
                                is_recursive=False):
        """Format any type of object & do so recursively if it's an iterable.

        Interpolates strings from the input args & kwargs.

        This is not a full on deepcopy, and it's on purpose not a full on
        deepcopy. It will handle dict, list, set, tuple for iteration, without
        any especial cuteness for other types or types not derived from these.

        A formattable object is a string or a special_type.

        special_type will call .get_value(kwargs) on the instance instead of
        formatting via the standard formatter.

        For iterable types, iterates recursively and formats all formattable
        objects in the iterable. Mappings will format both key and value.

        For any other type of object: returns it as is.

        This is what formatting or interpolating a string means:
        So where a string like this 'Piping {key1} the {key2} wild'
        And args/kwargs={'key1': 'down', 'key2': 'valleys', 'key3': 'value3'}

        Then this will return string: "Piping down the valleys wild"

        special_type will call .get_value(kwargs) on the instance instead of
        formatting via the standard formatter.

        Args:
            obj (any type): Recurse through me and format strings found in
                            iterables. Does not mutate the input object.
            args (iterable): Variable list positional args for formatting
                             expression.
            kwargs (iterable): Variable mapping of keyword args for formatting
                               expression.
            used_args (set): Mutating set of args used during formatting.
            memo (dict): Don't use. Used internally on recursion to optimize
                         recursive loops.
            is_recursive (bool): Don't use. Used internally to specify that
                                 in an active recursion loop.

        Returns:
            The input object formatted, if that object was a string or
            special_type. If the input obj was an iterable, an iterable
            identical in structure to the input, except where formatting
            changed a value from a string or special type to the formatting
            expression's evaluated value.

        """
        if memo is None:
            memo = {}

        obj_id = id(obj)
        already_done = memo.get(obj_id, None)
        if already_done is not None:
            return already_done

        # order important. passthrough supersedes special supersedes str.
        if self.passthrough_types and isinstance(obj, self.passthrough_types):
            new = obj
        elif self.special_types and isinstance(obj, self.special_types):
            new = obj.get_value(kwargs)
        elif isinstance(obj, str):
            new = self._format_keep_type(
                obj, args, kwargs, used_args,
                recursion_depth=self._FORMAT_SPEC_RECURSION_DEPTH,
                is_recursive=is_recursive)
        elif isinstance(obj, (bytes, bytearray)):
            new = obj
        elif isinstance(obj, Mapping):
            # dicts
            new = obj.__class__(
                (
                    self._get_formatted_iterable(
                        k, args, kwargs, used_args, memo, is_recursive),
                    self._get_formatted_iterable(
                        v, args, kwargs, used_args, memo, is_recursive)
                )
                for k, v in obj.items()
            )
        elif isinstance(obj, (Sequence, Set)):
            # list, set, tuple. Bytes and str won't fall into this branch coz
            # they're explicitly checked further up in the if.
            new = obj.__class__(
                self._get_formatted_iterable(
                    v, args, kwargs, used_args, memo, is_recursive)
                for v in obj)
        else:
            # int, float, bool, function, et.
            return obj

        # If is its own copy, don't memoize.
        if new is not obj:
            memo[obj_id] = new

        return new



"""pypyr step yaml definition classes - domain specific language."""
from functools import reduce
import logging
from pypyr.utils.filesystem import (ObjectRewriter,
                                    StreamRewriter)

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


class FileInRewriterStep():
    """A pypyr step that represents a file in-out rewriter.

    This models a step that takes config like this:
        root_key:
            in: str/path-like, or list of str/paths. Mandatory.
            out: str/path-like. Optional.

    The run_step method does the actual work.
    """

    def __init__(self, name, root_key, context):
        """Initialize the FileInRewriterStep.

        The step config in the context dict looks like this:
            root_key:
                in: str/path-like, or list of str/paths. Mandatory.
                out: str/path-like. Optional.
        Args:
            name: Unique name for step. Likely __name__ of calling step.
            root_key: str. Context key name where step's config is saved under.
            context: pypyr.context.Context. Look for config in this context
                     instance.

        """
        assert name, ("name parameter must exist for FileInRewriterStep.")
        assert root_key, ("root_key param must exist for FileInRewriterStep.")
        assert context, ("context param must exist for FileInRewriterStep.")
        # this way, logs output as the calling step, which makes more sense
        # to end-user than a mystery steps.dsl.blah logging output.
        self.logger = logging.getLogger(name)

        self.context = context
        root_dict = context[root_key]
        # this verifies both root_key and child exists
        context.assert_child_key_has_value(parent=root_key,
                                           child='in',
                                           caller=name)

        self.path_in = context.get_formatted_value(root_dict['in'])
        self.path_out = context.get_formatted_value(root_dict.get('out', None))

    def run_step(self, rewriter):
        """Do the file in to out rewrite.

        Doesn't do anything more crazy than call files_in_to_out on the
        rewriter.

        Args:
            rewriter: pypyr.filesystem.FileRewriter instance.
        """
        assert rewriter, ("FileRewriter instance required to run "
                          "FileInRewriterStep.")
        rewriter.files_in_to_out(in_path=self.path_in, out_path=self.path_out)


class ObjectRewriterStep(FileInRewriterStep):
    """A pypyr step that uses an object representer to rewrite files in out."""

    def run_step(self, representer):
        """Do the object in-out rewrite.

        Args:
            representer: A pypyr.filesystem.ObjectRepresenter instance.

        """
        assert representer, ("ObjectRepresenter instance required to run "
                             "ObjectRewriterStep.")
        rewriter = ObjectRewriter(self.context.get_formatted_value,
                                  representer)
        super().run_step(rewriter)


class StreamRewriterStep(FileInRewriterStep):
    """A pypyr step that uses a streaming file in-out rewriter.

    This models a step that takes config like this:
        root_key:
            in: str/path-like, or list of str/paths. Mandatory.
            out: str/path-like. Optional.
            replacePairs: mandatory. Dictionary where items are:
                          'find_string': 'replace_string'

    """

    def run_step(self):
        """Do the file in-out rewrite."""
        rewriter = StreamRewriter(self.context.iter_formatted_strings)
        super().run_step(rewriter)


class StreamReplacePairsRewriterStep(FileInRewriterStep):
    """A stream rewriter step that uses replacePairs."""

    def __init__(self, name, root_key, context):
        """Initialize the FileInRewriterStep.

        The step config in the context dict looks like this:
            root_key:
                in: str/path-like, or list of str/paths. Mandatory.
                out: str/path-like. Optional.
                replacePairs: mandatory. Dictionary where items are:
                              'find_string': 'replace_string'
        Args:
            name: Unique name for step. Likely __name__ of calling step.
            root_key: str. Context key name where step's config is saved under.
            context: pypyr.context.Context. Look for config in this context
                     instance.

        """
        super().__init__(name=name, root_key=root_key, context=context)
        # this verifies both root_key and child exists
        context.assert_child_key_has_value(parent=root_key,
                                           child='replacePairs',
                                           caller=name)

        self.replace_pairs = context[root_key]['replacePairs']

    def run_step(self):
        """Write in to out, replacing strings per the replace_pairs."""
        formatted_replacements = self.context.get_formatted_value(
            self.replace_pairs)

        iter = StreamReplacePairsRewriterStep.iter_replace_strings(
            formatted_replacements)
        rewriter = StreamRewriter(iter)
        super().run_step(rewriter)

    @staticmethod
    def iter_replace_strings(replacements):
        """Create a function that uses replacement pairs to process a string.

        The returned function takes an iterator and yields on each processed
        line.

        Args:
            replacements: Dict containing 'find_string': 'replace_string' pairs

        Returns:
            function with signature: iterator of strings = function(iterable)

        """
        def function_iter_replace_strings(iterable_strings):
            """Yield a formatted string from iterable_strings using a generator.

            Args:
                iterable_strings: Iterable containing strings. E.g a file-like
                                  object.

            Returns:
                Yields formatted line.

            """
            for string in iterable_strings:
                yield reduce((lambda s, kv: s.replace(*kv)),
                             replacements.items(),
                             string)

        return function_iter_replace_strings



"""Step that stops all pypyr execution."""
import logging
from pypyr.errors import Stop

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Immediately stop all processing."""
    logger.debug("started")

    logger.info("Stop: stopping pypyr...")
    raise Stop("pypyr.steps.stop stopped pypyr execution")



"""pypyr step that wipes the entire context."""
import logging

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Wipe the entire context.

    Args:
        Context is a dictionary or dictionary-like.
        Does not require any specific keys in context.
    """
    logger.debug("started")

    context.clear()
    logger.info("Context wiped. New context size: %s", len(context))

    logger.debug("done")



"""pypyr step that checks if a file or directory path exists."""
import logging
from pypyr.errors import KeyInContextHasNoValueError
import pypyr.utils.filesystem

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Check if a file or directory path exists.

    Args:
        context: pypyr.context.Context. Mandatory.
                 The following context key must exist
                - pathsToCheck. str/path-like or list of str/paths.
                                Path to file on disk to check.

    All inputs support formatting expressions. Supports globs.

    This step creates pathCheckOut in context, containing the results of the
    path check operation.

    pathCheckOut:
        'inpath':
            exists: true # bool. True if path exists.
            count: 0 # int. Number of files found for in path.
            found: ['path1', 'path2'] # list of strings. Paths of files found.

    [count] is 0 if no files found. If you specified a single input
    path to check and it exists, it's going to be 1. If you specified multiple
    in paths or a glob expression that found more than 1 result, well, take a
    guess.

    [found] is a list of all the paths found for the [inpath]. If you passed
    in a glob or globs, will contain the globs found for [inpath].

    This means you can do an existence evaluation like this in a formatting
    expression: '{pathCheckOut[inpathhere][exists]}'

    Returns:
        None. updates context arg.

    Raises:
        pypyr.errors.KeyNotInContextError: pathExists missing in context.
        pypyr.errors.KeyInContextHasNoValueError: pathCheck exists but is None.

    """
    logger.debug("started")
    context.assert_key_has_value(key='pathCheck', caller=__name__)

    paths_to_check = context['pathCheck']

    if not paths_to_check:
        raise KeyInContextHasNoValueError("context['pathCheck'] must have a "
                                          f"value for {__name__}.")

    # pathsToCheck can be a string or a list in case there are multiple paths
    if isinstance(paths_to_check, list):
        check_me = paths_to_check
    else:
        # assuming it's a str/path at this point
        check_me = [paths_to_check]

    out = {}
    total_found = 0

    for path in check_me:
        logger.debug("checking path: %s", path)
        formatted_path = context.get_formatted_value(path)
        found_paths = pypyr.utils.filesystem.get_glob(formatted_path)
        no_of_paths = len(found_paths)
        out[path] = {
            'exists': no_of_paths > 0,
            'count': no_of_paths,
            'found': found_paths
        }
        total_found = total_found + no_of_paths

    context['pathCheckOut'] = out

    logger.info(f'checked {len(out)} path(s) and found {total_found}')
    logger.debug("done")



"""pypyr step copies context values from already existing context values.

This is handy if you need to prepare certain keys in context where a next step
might need a specific key. If you already have the value in context, you can
create a new key (or update existing key) with that value.

So let's say you already have context['currentKey'] = 'eggs'.
If you run newKey: currentKey, you'll end up with context['newKey'] == 'eggs'
"""
import logging

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Create/copy new context keys from already existing context keys.

    Context is a dictionary or dictionary-like.
    context['contextCopy'] must exist. It's a dictionary.
    Will iterate context['contextCopy'] and save the values as new keys to the
    context.

    For example, say input context is:
        key1: value1
        key2: value2
        key3: value3
        contextCopy:
            key2: key1
            key4: key3

    This will result in return context:
        key1: value1
        key2: value1
        key3: value3
        key4: value3
    """
    logger.debug("started")
    context.assert_key_has_value(key='contextCopy', caller=__name__)

    for k, v in context['contextCopy'].items():
        logger.debug("copying context %s to value from context %s", k, v)
        context[k] = context[v]

    logger.debug("done")



"""Context parser that returns a dictionary from input args.

Takes input args and returns a dictionary where each element
becomes the key, with value to true.

So cli input like this "ham eggs bacon", will yield:
{'ham': True, 'eggs': True, 'bacon': True}
"""
import logging

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)


def get_parsed_context(args):
    """Parse input context string and returns context as dictionary."""
    logger.debug("starting")
    if not args:
        logger.debug("pipeline invoked without context arg set. For "
                     "this keys parser you're looking for something "
                     "like: pypyr pipelinename spam eggs \n"
                     "or: pypyr pipelinename spam."
                     )
        return None

    # for each arg, project (element-name, true)
    return dict((element, True) for element in args)



"""pypyr step that parses json file for string substitutions, writes output."""
import logging
from pypyr.steps.dsl.fileinoutrewriter import ObjectRewriterStep
from pypyr.utils.filesystem import JsonRepresenter

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Parse input json file and substitute {tokens} from context.

    Loads json into memory to do parsing, so be aware of big files.

    Args:
        context: pypyr.context.Context. Mandatory.
                - fileFormatJson
                    - in. mandatory.
                      str, path-like, or an iterable (list/tuple) of
                      strings/paths. Each str/path can be a glob, relative or
                      absolute path.
                    - out. optional. path-like.
                      Can refer to a file or a directory.
                      will create directory structure if it doesn't exist. If
                      in-path refers to >1 file (e.g it's a glob or list), out
                      path can only be a directory - it doesn't make sense to
                      write >1 file to the same single file (this is not an
                      appender.) To ensure out_path is read as a directory and
                      not a file, be sure to have the path separator (/) at the
                      end.
                      If out_path is not specified or None, will in-place edit
                      and overwrite the in-files.

    Returns:
        None.

    Raises:
        FileNotFoundError: take a guess
        pypyr.errors.KeyNotInContextError: fileFormatJson or
            fileFormatJson['in'] missing in context.
        pypyr.errors.KeyInContextHasNoValueError: fileFormatJson or
            fileFormatJson['in'] exists but is None.

    """
    logger.debug("started")

    ObjectRewriterStep(__name__, 'fileFormatJson', context).run_step(
        JsonRepresenter())

    logger.debug("done")



"""Global cache for context_parser modules.

Attributes:
    contextparser_cache: global instance of the context_parser cache.
                      Use this attribute to access the cache from elsewhere.
"""
import logging
from pypyr.cache.cache import Cache
import pypyr.moduleloader

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)


class ContextParserCache(Cache):
    """Get functions from the pypeloader cache."""

    def get_context_parser(self, parser_module_name):
        """Get cached get_parsed_context function. Adds to cache if not exist.

        Args:
            parser_module_name: load the module specified by this, get its
                                get_parsed_context function and add it to
                                cache.

        Return:
            function: the get_parsed_context function of module specified
                      by parser_module_name.
        """
        logger.debug("starting")

        parser_function = self.get(parser_module_name,
                                   lambda: load_the_parser(parser_module_name))

        logger.debug("done")
        return parser_function


# global instance of the cache. use this to access the cache from elsewhere.
contextparser_cache = ContextParserCache()


def load_the_parser(parser_module_name):
    """Load the module specified by parser_module_name, returns its function.

    Args:
        parser_module_name: string: name of module to load

    Returns:
        function: the get_parsed_context function in the module
    """
    logger.debug("starting")

    parser_module = pypyr.moduleloader.get_module(parser_module_name)
    logger.debug("context parser module found: %s", parser_module_name)

    try:
        get_parsed_context = getattr(
            parser_module, 'get_parsed_context'
        )
    except AttributeError:
        logger.error(
            "The context parser %s doesn't have a "
            "get_parsed_context(context_arg): "
            "function.",
            parser_module_name
        )
        raise

    logger.debug("done")
    return get_parsed_context



"""Step that stops current pipeline."""
import logging
from pypyr.errors import StopPipeline

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Stop current pipeline."""
    logger.debug("started")

    logger.info("StopPipeline: stopping pipeline...")
    raise StopPipeline("pypyr.steps.stoppipeline stopped current pipeline")



"""pypyr steps runner.

pipelinerunner uses this to parse and run steps.
"""

import logging
from pypyr.dsl import Step
from pypyr.errors import (ControlOfFlowInstruction,
                          Jump,
                          Stop,
                          StopStepGroup)

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)


class StepsRunner():
    """Run step-groups and steps.

    If you're wanting to run steps just like the pypyr cli does,
    run_step_groups() is a sensible entrypoint.
    """

    def __init__(self, pipeline_definition, context):
        """Initialize the Step Runner with the pipeline to maintain state.

        Args:
            pipeline_definition: pipeline yaml
            context: pypyr.context.Context. The pypyr context. Will mutate.
        """
        self.context = context
        self.pipeline = pipeline_definition

    def get_pipeline_steps(self, step_group):
        """Get the specified step-group's step from the pipeline.

        If there is no steps_group sequence on the pipeline, return None.
        Guess you could theoretically want to run a pipeline with nothing in
        it.

        Args:
            step_group: (str) Name of step-group

        Returns:
            Iterable collection of steps in the step-group.
        """
        logger.debug("starting")
        assert step_group

        logger.debug("retrieving %s steps from pipeline", step_group)
        if step_group in self.pipeline:
            steps = self.pipeline[step_group]

            if steps is None:
                logger.warning(
                    "%s: sequence has no elements. So it won't do anything.",
                    step_group,
                )
                logger.debug("done")
                return None

            steps_count = len(steps)

            logger.debug("%s steps found under %s in pipeline definition.",
                         steps_count, step_group)

            logger.debug("done")
            return steps
        else:
            logger.debug(
                "pipeline doesn't have a %(steps_group)s collection. Add a "
                "%(steps_group)s: sequence to the yaml if you want "
                "%(steps_group)s actually to do something.",
                {"steps_group": step_group}
            )
            logger.debug("done")
            return None

    def run_failure_step_group(self, group_name):
        """Run the group_name if it exists, as a failure handler..

        This function will swallow all errors, to prevent obfuscating the error
        condition that got it here to begin with.
        """
        logger.debug("starting")
        try:
            # if no group_name exists, it'll do nothing.
            self.run_step_group(group_name, raise_stop=True)
        except Stop:
            logger.debug("Stop instruction: done with failure handler %s.",
                         group_name)
            raise
        except Exception as exception:
            logger.error("Failure handler also failed. Swallowing.")
            logger.error(exception)

        logger.debug("done")

    def run_pipeline_steps(self, steps):
        """Run the run_step(context) method of each step in steps.

        Args:
            steps: list. Sequence of Steps to execute
        """
        logger.debug("starting")
        assert isinstance(self.context, dict), (
            "context must be a dictionary, even if empty {}.")

        if steps is None:
            logger.debug("No steps found to execute.")
        else:
            step_count = 0

            for step in steps:
                step_instance = Step(step, self)
                step_instance.run_step(self.context)
                step_count += 1

            logger.debug("executed %s steps", step_count)

        logger.debug("done")

    def run_step_group(self, step_group_name, raise_stop=False):
        """Get the specified step group from the pipeline and run its steps."""
        logger.debug("starting %s", step_group_name)
        assert step_group_name

        steps = self.get_pipeline_steps(step_group=step_group_name)

        try:
            self.run_pipeline_steps(steps=steps)
        except Jump as jump:
            logger.debug("jump: jumping to %s", jump.groups)
            self.run_step_groups(groups=jump.groups,
                                 success_group=jump.success_group,
                                 failure_group=jump.failure_group)
            logger.debug("jump: done jumping to %s", jump.groups)
        except StopStepGroup:
            logger.debug("StopStepGroup: stopped %s", step_group_name)
            if raise_stop:
                raise

        logger.debug("done %s", step_group_name)

    def run_step_groups(self, groups, success_group, failure_group):
        """Run stepgroups specified, with the success and failure handlers.

        Args:
            groups: (list) list of step-group names to run.
            success_group: (str) name of group to run on successful completion
                           of groups.
            failure_group: (str) name of group to run on error

        Returns:
            None
        """
        logger.debug("starting")

        if not groups:
            raise ValueError("you must specify which step-groups you want to "
                             "run. groups is None.")
        try:
            # run main steps
            for step_group in groups:
                self.run_step_group(step_group)

            # if nothing went wrong, run on_success
            if success_group:
                logger.debug(
                    "pipeline steps complete. Running %s steps now.",
                    success_group)
                self.run_step_group(success_group)
            else:
                logger.debug(
                    "pipeline steps complete. No success group specified.")
        except (ControlOfFlowInstruction, Stop):
            # Control-of-Flow/Stop are instructions to go somewhere
            # else, not errors per se.
            raise
        except Exception:
            # yes, yes, don't catch Exception. Have to, though, to run failure
            # handler. Also, it does raise it back up.
            do_raise = True
            if failure_group:
                logger.error(
                    "Something went wrong. Will now try to run %s.",
                    failure_group)

                # failure_step_group will log but swallow any errors except
                # Stop. This so that pipeline can quit failure_handler
                # via stop without raising an error.
                try:
                    self.run_failure_step_group(failure_group)
                except StopStepGroup:
                    do_raise = False
            else:
                logger.debug(
                    "Something went wrong. No failure group specified.")

            if do_raise:
                logger.debug("Raising original exception to caller.")
                raise

        logger.debug("done")



"""pypyr step that sets context values from formatting expressions.

This is handy if you need to prepare certain keys in context where a next step
might need a specific key. If you already have the value in context, you can
create a new key (or update existing key) with that value.
"""
import logging

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Set new context keys from formatting expressions with substitutions.

    Context is a dictionary or dictionary-like.
    context['contextSetf'] must exist. It's a dictionary.
    Will iterate context['contextSetf'] and save the values as new keys to the
    context.

    For example, say input context is:
        key1: value1
        key2: value2
        key3: value3
        contextSetf:
            key2: 'aaa_{key1}_zzz'
            key4: 'bbb_{key3}_yyy'

    This will result in return context:
        key1: value1
        key2: aaa_value1_zzz
        key3: bbb_value3_yyy
        key4: value3
    """
    logger.debug("started")
    context.assert_key_has_value(key='contextSetf', caller=__name__)

    for k, v in context['contextSetf'].items():
        logger.debug("setting context %s to value from context %s", k, v)
        context[context.get_formatted_value(
            k)] = context.get_formatted_value(v)

    logger.info("set %d context items.", len(context['contextSetf']))

    logger.debug("done")



"""Guard utility functions for ensuring things exist or have values."""
from pypyr.errors import (ContextError,
                          KeyNotInContextError,
                          KeyInContextHasNoValueError)


def assert_key_exists(obj, key, caller, parent=None):
    """Assert that object contains key.

    Error messages are structured as if obj is a pypyr Context.

    Args:
        obj (mapping): object to check for key.
        key (any valid key type): validates that this key exists in context
        caller: string. calling function or module name - this used to
                construct error messages. Tip: use .__name__
        parent (any valid key type): parent key name. Used to construct error
                                     messages to indicate the name of missing
                                     obj in context.

    Raises:
        KeyNotInContextError: When key doesn't exist in context.

    """
    try:
        if key not in obj:
            if parent:
                msg = (f"context[{parent!r}][{key!r}] doesn't "
                       f"exist. It must exist for {caller}.")
            else:
                msg = (
                    f"context[{key!r}] doesn't exist. "
                    f"It must exist for {caller}.")

            raise KeyNotInContextError(msg)
    except TypeError as err:
        # catches None on obj or obj not iterable
        if parent:
            msg = (f"context[{parent!r}] must exist, be iterable and contain "
                   f"{key!r} for {caller}. {err}")
        else:
            msg = (f"context[{key!r}] must exist and be iterable for "
                   f"{caller}. {err}")
        raise ContextError(msg) from err


def assert_key_has_value(obj, key, caller, parent=None):
    """Assert key exists in obj and has value.

    Does not eval on truthy. Empty string or 0 counts as a value.

    Error messages are structured as if obj is a pypyr Context.

    Args:
        obj (mapping): object to check for key.
        key (any valid key type): key to check in obj
        caller (string): calling function name - this used to construct
                         error messages. Tip: use .__name__
        parent (string): parent key name. Used to construct error messages to
                         indicate the name of obj in context.

    Raises:
        ContextError: if obj is None or not iterable.
        KeyInContextHasNoValueError: context[key] is None
        KeyNotInContextError: Key doesn't exist
    """
    assert_key_exists(obj, key, caller, parent)
    if obj[key] is None:
        if parent:
            msg = (f"context[{parent!r}][{key!r}] must have a value for "
                   f"{caller}.")
        else:
            msg = f"context[{key!r}] must have a value for {caller}."

        raise KeyInContextHasNoValueError(msg)



"""Step that asserts something is true or equal to something else."""
import logging
from pypyr.errors import KeyNotInContextError
from pypyr.utils.types import cast_to_bool

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Assert that something is True or equal to something else.

    Args:
        context (pypyr.context.Context): context is mandatory.
        Uses the following context keys in context:
            - assert
                - this (any): mandatory. If assert['equals'] not specified,
                  eval as boolean.
                - equals (any): optional. Any type.

    Takes one of three input forms:
        assert: evaluate me

        or

        assert:
            this: evaluate me

        or

        assert:
            this: compare me
            equals: to this

    If context['assert'] is not a dict, evaluate contents as bool.
    If context['assert'] is a dict:
        - If context['assert']['this'] evaluates to False raises error.
        - If context['assert']['equals'] exists, raises error if assert.this
          != assert.equals.

    All input forms support string substitutions.

    Returns:
        None

    Raises:
        AssertionError: if assert evaluates to False.

    """
    logger.debug("started")
    assert context, f"context must have value for {__name__}"

    context.assert_key_exists('assert', __name__)

    assert_context = context.get_formatted('assert')

    is_equals_there = False
    is_this_there = False
    if isinstance(assert_context, dict):
        is_this_there = 'this' in assert_context
        if is_this_there:
            assert_this = assert_context['this']
        else:
            assert_this = assert_context

        is_equals_there = 'equals' in assert_context
        if is_equals_there:
            if not is_this_there:
                raise KeyNotInContextError(
                    "you have to set assert.this to use assert.equals.")
            assert_equals = assert_context['equals']
            # compare assertThis to assertEquals
            logger.debug("comparing assert['this'] to assert['equals'].")
            assert_result = (assert_this == assert_equals)
        else:
            # nothing to compare means treat assert or assert.this as a bool.
            if is_this_there:
                logger.debug("evaluating assert['this'] as a boolean.")
            else:
                logger.debug("assert is a dict but contains no `this`. "
                             "evaluating assert value as a boolean.")
            assert_result = cast_to_bool(assert_this)
    else:
        # assert key has a non-dict value so eval directly as a bool.
        logger.debug("evaluating assert value as a boolean.")
        assert_result = cast_to_bool(assert_context)

    logger.info("assert evaluated to %s", assert_result)

    if not assert_result:
        if is_equals_there:
            # emit type to help user, but not the actual field contents.
            type_this = type(assert_this).__name__
            type_equals = type(assert_equals).__name__
            error_text = (
                f"assert assert['this'] is of type {type_this} "
                f"and does not equal assert['equals'] of type {type_equals}.")
        else:
            og_assert = (context['assert']['this']
                         if is_this_there else context['assert'])
            # original literal hard-coded in pipe, so presumably not a
            # sensitive value.
            error_text = (
                f"assert {og_assert} evaluated to False.")
        raise AssertionError(error_text)

    logger.debug("done")



"""pypyr step saves the current local datetime to context."""
from datetime import datetime
from dateutil.tz import tzlocal
import logging

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Save current local datetime to context.

    Args:
        context: pypyr.context.Context. Mandatory.
                 The following context key is optional:
                - nowIn. str. Datetime formatting expression. For full list of
                  possible expressions, check here:
                  https://docs.python.org/3.7/library/datetime.html#strftime-and-strptime-behavior

    All inputs support pypyr formatting expressions.

    This step creates now in context, containing a string representation of the
    timestamp. If input formatting not specified, defaults to ISO8601.

    Default is:
    YYYY-MM-DDTHH:MM:SS.ffffff, or, if microsecond is 0, YYYY-MM-DDTHH:MM:SS

    Returns:
        None. updates context arg.

    """
    logger.debug("started")

    format_expression = context.get_formatted_value(context.get('nowIn', None))

    if format_expression:
        context['now'] = datetime.now(tzlocal()).strftime(format_expression)
    else:
        context['now'] = datetime.now(tzlocal()).isoformat()

    logger.info("timestamp %s saved to context now", context['now'])
    logger.debug("done")



"""pypyr step that executes a cmd as a sub-process.

You cannot use things like exit, return, shell pipes, filename wildcards,
environment,variable expansion, and expansion of ~ to a user’s home
directory.
"""
import logging
from pypyr.steps.dsl.cmd import CmdStep

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Run command, program or executable.

    Context is a dictionary or dictionary-like.

    Context must contain the following keys:
    cmd: <<cmd string>> (command + args to execute.)

    OR, as a dict
    cmd:
        run: str. mandatory. <<cmd string>> command + args to execute.
        save: bool. defaults False. save output to cmdOut.

    Will execute the command string in the shell as a sub-process.
    Escape curly braces: if you want a literal curly brace, double it like
    {{ or }}.

    If save is True, will save the output to context as follows:
        cmdOut:
            returncode: 0
            stdout: 'stdout str here. None if empty.'
            stderr: 'stderr str here. None if empty.'

    cmdOut.returncode is the exit status of the called process. Typically 0
    means OK. A negative value -N indicates that the child was terminated by
    signal N (POSIX only).

    context['cmd'] will interpolate anything in curly braces for values
    found in context. So if your context looks like this:
        key1: value1
        key2: value2
        cmd: mything --arg1 {key1}

    The cmd passed to the shell will be "mything --arg value1"
    """
    logger.debug("started")

    CmdStep(name=__name__, context=context).run_step(is_shell=False)

    logger.debug("done")



"""Context parser that adds the input argument string into the pypyr context.

Takes any arbitrary string and adds it to the context dict as argString.

So a string like this "ham eggs bacon", will yield context:
{'argString': 'ham eggs bacon'}
"""
import logging

# getLogger will grab the parent logger context, so your loglevel and
# formatting will inherit correctly automatically from the pypyr core.
logger = logging.getLogger(__name__)


def get_parsed_context(args):
    """Parse input context string and returns context as dictionary."""
    logger.debug("starting")
    if not args:
        logger.debug("pipeline invoked without context arg set. For "
                     "this string parser you're looking for something "
                     "like: pypyr pipelinename spam and eggs"
                     )
        return {'argString': None}

    # the list that's parsed from the input args is named argList
    return dict({'argString': ' '.join(args)})



"""Utility functions for evaluating expressions."""


def eval_string(input_string, locals):
    """Dynamically evaluates the input_string expression.

    This provides dynamic python eval of an input expression. The return is
    whatever the result of the expression is.

    Use with caution: since input_string executes any arbitrary code object,
    the potential for damage is great.

    For reasons best known to eval(), it EAFPs the locals() look-up 1st and if
    it raises a KeyNotFound error, moves on to globals.

    This means if you pass in a custom dict type for locals that does not
    raise a KeyNotFound or derived exception,  and thus eval doesn't work.

    Therefore, in order to be used as the locals arg, if you use a dict with
    a customized exception when look-up fails that is not an instance of
    KeyNotFound, you should initialize to a new dict like this:
        out = eval_string('1==1', dict(mydict))

    Args:
        input_string: str. expression to evaluate.
        locals: dict-like. Mapping object containing scope for eval.

    Returns:
        Whatever object results from the string expression valuation.

    """
    # empty globals arg will append __builtins__ by default
    return eval(input_string, {}, locals)



"""Context parser that returns a dictionary from a local yaml file."""

from collections.abc import MutableMapping
import logging
import ruamel.yaml as yaml

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)


def get_parsed_context(args):
    """Parse input as path to a yaml file, returns context as dictionary."""
    logger.debug("starting")
    if not args:
        raise AssertionError(
            "pipeline must be invoked with context arg set. For "
            "this yaml parser you're looking for something like:\n"
            "pypyr pipelinename ./myyamlfile.yaml")
    path = ' '.join(args)
    logger.debug("attempting to open file: %s", path)
    with open(path) as yaml_file:
        yaml_loader = yaml.YAML(typ='safe', pure=True)
        payload = yaml_loader.load(yaml_file)

    if not isinstance(payload, MutableMapping):
        raise TypeError("yaml input should describe a dictionary at the top "
                        "level. You should have something like\n"
                        "key1: value1\n key2: value2\n"
                        "in the yaml top-level, not \n- value1\n- value2")

    logger.debug("yaml file parsed. Count: %d", len(payload))

    logger.debug("done")
    return payload



"""Step that gets, sets and unsets env vars."""
import os
import logging
from pypyr.errors import KeyNotInContextError

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Get, set, unset $ENVs.

    Context is a dictionary or dictionary-like. context is mandatory.

    Input context is:
        env:
            get: {dict}
            set: {dict}
            unset: [list]

    At least one of env's sub-keys (get, set or unset) must exist.

    This step will run whatever combination of Get, Set and Unset you specify.
    Regardless of combination, execution order is Get, Set, Unset.
    """
    logger.debug("started")
    assert context, f"context must have value for {__name__}"

    context.assert_key_has_value('env', __name__)

    found_get = env_get(context)
    found_set = env_set(context)
    found_unset = env_unset(context)

    # at least 1 of envGet, envSet or envUnset must exist in context
    if not (found_get or found_set or found_unset):
        raise KeyNotInContextError(
            "context must contain any combination of "
            "env['get'], env['set'] or env['unset'] for "
            f"{__name__}")

    logger.debug("done")


def env_get(context):
    """Get $ENVs into the pypyr context.

    Context is a dictionary or dictionary-like. context is mandatory.

    context['env']['get'] must exist. It's a dictionary.
    Values are the names of the $ENVs to write to the pypyr context.
    Keys are the pypyr context item to which to write the $ENV values.

    For example, say input context is:
        key1: value1
        key2: value2
        pypyrCurrentDir: value3
        env:
            get:
                pypyrUser: USER
                pypyrCurrentDir: PWD

    This will result in context:
        key1: value1
        key2: value2
        key3: value3
        pypyrUser: <<value of $USER here>>
        pypyrCurrentDir: <<value of $PWD here, not value3>>
    """
    get = context.get_formatted_value(context['env'].get('get', None))

    exists = False
    if get:
        logger.debug("start")

        for k, v in get.items():
            logger.debug("setting context %s to $ENV %s", k, v)
            context[k] = os.environ[v]

        logger.info("saved %d $ENVs to context.", len(get))
        exists = True

        logger.debug("done")
    return exists


def env_set(context):
    """Set $ENVs to specified string. from the pypyr context.

    Args:
        context: is dictionary-like. context is mandatory.
                 context['env']['set'] must exist. It's a dictionary.
                 Values are strings to write to $ENV.
                 Keys are the names of the $ENV values to which to write.

    For example, say input context is:
        key1: value1
        key2: value2
        key3: value3
        env:
            set:
                MYVAR1: {key1}
                MYVAR2: before_{key3}_after
                MYVAR3: arbtexthere

    This will result in the following $ENVs:
    $MYVAR1 = value1
    $MYVAR2 = before_value3_after
    $MYVAR3 = arbtexthere

    Note that the $ENVs are not persisted system-wide, they only exist for
    pypyr sub-processes, and as such for the following steps during this pypyr
    pipeline execution. If you set an $ENV here, don't expect to see it in your
    system environment variables after the pipeline finishes running.
    """
    env_set = context.get_formatted_value(context['env'].get('set', None))

    exists = False
    if env_set:
        logger.debug("started")

        for k, v in env_set.items():
            logger.debug("setting $%s", k)
            os.environ[k] = v

        logger.info("set %s $ENVs from context.", len(env_set))
        exists = True

        logger.debug("done")

    return exists


def env_unset(context):
    """Unset $ENVs.

    Context is a dictionary or dictionary-like. context is mandatory.

    context['env']['unset'] must exist. It's a list.
    List items are the names of the $ENV values to unset.

    For example, say input context is:
        key1: value1
        key2: value2
        key3: value3
        env:
            unset:
                MYVAR1
                MYVAR2

    This will result in the following $ENVs being unset:
    $MYVAR1
    $MYVAR2
    """
    unset = context.get_formatted_value(context['env'].get('unset', None))

    exists = False
    if unset:
        logger.debug("started")

        for env_var_name in unset:
            logger.debug("unsetting $%s", env_var_name)
            try:
                del os.environ[env_var_name]
            except KeyError:
                # If user is trying to get rid of the $ENV, if it doesn't
                # exist, no real point in throwing up an error that the thing
                # you're trying to be rid off isn't there anyway.
                logger.debug(
                    "$%s doesn't exist anyway. As you were.", env_var_name)

        logger.info("unset %d $ENVs.", len(unset))
        exists = True

        logger.debug("done")
    return exists



"""pypyr step that executes a shell as a sub-process.

The command string must be formatted exactly as it would be when typed
at the shell prompt. This includes, for example, quoting or backslash escaping
filenames with spaces in them. The shell defaults to /bin/sh.
"""
import logging
from pypyr.steps.dsl.cmd import CmdStep

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Run shell command without shell interpolation.

    Context is a dictionary or dictionary-like.

    Context must contain the following keys:
    cmd: <<cmd string>> (command + args to execute.)

    OR, as a dict
    cmd:
        run: str. mandatory. <<cmd string>> command + args to execute.
        save: bool. defaults False. save output to cmdOut.

    Will execute command string in the shell as a sub-process.
    The shell defaults to /bin/sh.
    The context['cmd'] string must be formatted exactly as it would be when
    typed at the shell prompt. This includes, for example, quoting or backslash
    escaping filenames with spaces in them.
    There is an exception to this: Escape curly braces: if you want a literal
    curly brace, double it like {{ or }}.

    If save is True, will save the output to context as follows:
        cmdOut:
            returncode: 0
            stdout: 'stdout str here. None if empty.'
            stderr: 'stderr str here. None if empty.'

    cmdOut.returncode is the exit status of the called process. Typically 0
    means OK. A negative value -N indicates that the child was terminated by
    signal N (POSIX only).

    context['cmd'] will interpolate anything in curly braces for values
    found in context. So if your context looks like this:
        key1: value1
        key2: value2
        cmd: mything --arg1 {key1}

    The cmd passed to the shell will be "mything --arg value1"
    """
    logger.debug("started")

    CmdStep(name=__name__, context=context).run_step(is_shell=True)

    logger.debug("done")



"""Global cache for pypeloader modules.

Attributes:
    pypeloader_cache: global instance of the loader cache.
                      Use this attribute to access the cache from elsewhere.
"""
import logging
from pypyr.cache.cache import Cache
import pypyr.moduleloader

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)


class PypeLoaderCache(Cache):
    """Get functions from the pypeloader cache."""

    def get_pype_loader(self, loader=None):
        """Get cached get_pipeline_definition function.

        Defaults to pypyr.pypeloaders.fileloader if loader not specified.

        Args:
            loader: load the module specified by this, get its
                    get_pipeline_definition function and add it to cache.

        Return:
            function: the get_pipeline_definition function of module specified
                      by loader.
        """
        logger.debug("starting")

        if loader:
            logger.debug("you set the pype loader to: %s", loader)
        else:
            loader = 'pypyr.pypeloaders.fileloader'
            logger.debug("use default pype loader: %s", loader)

        loader_function = self.get(loader, lambda: load_the_loader(loader))

        logger.debug("done")
        return loader_function


# global instance of the cache. use this to access the cache from elsewhere.
pypeloader_cache = PypeLoaderCache()


def load_the_loader(loader):
    """Load the module specified by loader and returns its function.

    Args:
        loader: string: name of module to load

    Returns:
        function: the get_pipeline_definition function in the module
    """
    logger.debug("starting")

    # pipeline loading deliberately outside of try catch. The try catch
    # will try to run a failure-handler from the pipeline, but if the
    # pipeline doesn't exist there is no failure handler that can possibly
    # run so this is very much a fatal stop error.
    loader_module = pypyr.moduleloader.get_module(loader)

    try:
        get_pipeline_definition = getattr(
            loader_module, 'get_pipeline_definition'
        )
    except AttributeError:
        logger.error(
            "The pipeline loader %s doesn't have a "
            "get_pipeline_definition(pipeline_name, working_dir) "
            "function.",
            loader_module
        )
        raise

    logger.debug("loading the pipeline definition with %s", loader_module)
    logger.debug("%s done", loader_module)
    return get_pipeline_definition



"""pypyr step that clears context values.

This is handy if you run the same step multiple times in a pipeline and you
don't want the previously assigned context settings for that step to impact the
next iteration.
"""
import logging

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Remove specified keys from context.

    Args:
        Context is a dictionary or dictionary-like.
        context['contextClear'] must exist. It's a dictionary.
        Will iterate context['contextClear'] and remove those keys from
        context.

    For example, say input context is:
        key1: value1
        key2: value2
        key3: value3
        key4: value4
        contextClear:
            - key2
            - key4
            - contextClear

    This will result in return context:
        key1: value1
        key3: value3
    """
    logger.debug("started")
    context.assert_key_has_value(key='contextClear', caller=__name__)

    for k in context['contextClear']:
        logger.debug("removing %s from context", k)
        # slightly unorthodox pop returning None means you don't get a KeyError
        # if key doesn't exist
        context.pop(k, None)
        logger.info("removed %s from context", k)

    logger.debug("done")



"""Context parser that returns a dictionary from a key-value pair string.

Takes list of key=value pair string and returns a dictionary where
each pair becomes a dictionary element.

Don't have spaces in your values unless your really mean it. "k1=v1 ' k2'=v2"
will result in a context key name of ' k2' not 'k2'.

So cli input like this "pig=ham hen=eggs yummypig=bacon", will yield:
{'pig': 'ham', 'hen': ''eggs', 'yummypig': 'bacon'}
"""
import logging

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)


def get_parsed_context(args):
    """Parse input context args and returns context as dictionary."""
    logger.debug("starting")
    if not args:
        logger.debug("pipeline invoked without context arg set. For "
                     "this keyvaluepairs parser you're looking for "
                     "something like:\n"
                     "pypyr pipelinename key1=value1 key2=value2")
        return None

    # for each arg, project key=value
    return dict(element.split('=') for element in args)



"""Utility functions for polling."""
import time
import logging

# pypyr logger means the log level will be set correctly and output formatted.
logger = logging.getLogger(__name__)


def wait_until_true(interval, max_attempts):
    """Execute a decorated function until it returns True.

    Executes wrapped function at every number of seconds specified by interval,
    until wrapped function either returns True or max_attempts are exhausted,
    whichever comes 1st. The wrapped function can have any given signature.

    Use me if you always want to time out at max_attempts and you don't care
    about the while loop position counter value. If you do care, use
    while_until_true instead.

    Args:
        interval: In seconds. How long to wait between executing the wrapped
                  function.
        max_attempts: int. Execute wrapped function up to this limit.

    Returns:
        Bool. True if wrapped function returned True. False if reached
              max_attempts without the wrapped function ever returning True.
    """
    def decorator(f):
        logger.debug("started")

        def sleep_looper(*args, **kwargs):
            logger.debug("Looping every %s seconds for %s attempts",
                         interval, max_attempts)
            for i in range(1, max_attempts + 1):
                result = f(*args, **kwargs)
                if result:
                    logger.debug("iteration %s. Desired state reached.", i)
                    return True
                if i < max_attempts:
                    logger.debug("iteration %s. Still waiting. . .", i)
                    time.sleep(interval)
            logger.debug("done")
            return False
        return sleep_looper

    return decorator


def while_until_true(interval, max_attempts):
    """Execute a decorated function until it returns True.

    Executes wrapped function at every number of seconds specified by interval,
    until wrapped function either returns True or max_attempts are exhausted,
    whichever comes 1st.

    The difference between while_until_true and wait_until_true is that the
    latter will always loop to a max_attempts, whereas while_until_true will
    keep going indefinitely.

    The other notable difference to wait_until_true is that the wrapped
    function signature must be:
    func(counter, *args, **kwargs)

    This is because this decorator injects the while loop counter into the
    invoked function.

    Args:
        interval: In seconds. How long to wait between executing the wrapped
                  function.
        max_attempts: int. Execute wrapped function up to this limit. None
                      means infinite (or until wrapped function returns True).
                      Passing anything <0 also means infinite.

    Returns:
        Bool. True if wrapped function returned True. False if reached
              max_attempts without the wrapped function ever returning True.
    """
    def decorator(f):
        logger.debug("started")

        def sleep_looper(*args, **kwargs):
            if max_attempts:
                logger.debug("Looping every %s seconds for %s attempts",
                             interval, max_attempts)
            else:
                logger.debug("Looping every %s seconds.", interval)

            i = 0
            result = False

            # pragma for coverage: cov can't figure out the branch construct
            # with the dynamic function invocation, it seems, so marks the
            # branch partial. unit test cov is 100%, though.
            while not result:  # pragma: no branch
                i += 1
                result = f(i, *args, **kwargs)
                if result:
                    logger.debug("iteration %s. Desired state reached.", i)
                    break
                elif max_attempts:
                    if i < max_attempts:
                        logger.debug("iteration %s. Still waiting. . .", i)
                        time.sleep(interval)
                    else:
                        logger.debug("iteration %s. Max attempts exhausted.",
                                     i)
                        break
                else:
                    # result False AND max_attempts is None means keep looping
                    # because None = infinite
                    logger.debug("iteration %s. Still waiting. . .", i)
                    time.sleep(interval)
            logger.debug("done")
            return result

        return sleep_looper

    return decorator



"""Utility functions for type checking, conversion and handling."""


def are_all_this_type(type, *objects):
    """Return True if all objects are instances of this type.

    Args:
        type: Type to check against, e.g str, dict
        objects: *args of objects. Each of these will be checked against type.

    Returns:
        True is all objects are instances of specified Type.
        False if any or all of the objects are not instances of Type.

    """
    # return true if all objects are instances of type
    return all(isinstance(object, type) for object in objects)


def cast_to_bool(obj):
    """Convert to bool with special string parsing.

    If obj is string: case-insensitive 'True', '1' or '1.0' is True.
    It will be False for all other strings.

    Any other object evaluates with standard truthy.

    Args:
        obj (any): Get truthy from this input object

    Returns:
        bool. Truthy value of input obj.
    """
    if isinstance(obj, str):
        return obj.lower() in ['true', '1', '1.0']
    else:
        return bool(obj)


def cast_to_type(obj, out_type):
    """Cast obj to out_type if it's not out_type already.

    If the obj happens to be out_type already, it just returns obj as is.

    Args:
        obj: input object
        out_type: type.

    Returns:
        obj cast to out_type. Usual python conversion / casting rules apply.

    """
    in_type = type(obj)
    if out_type is in_type:
        # no need to cast.
        return obj
    else:
        return out_type(obj)



"""Global cache for run_step functions of Steps.

Attributes:
    step_cache: global instance of the context_parser cache.
                      Use this attribute to access the cache from elsewhere.
"""
import logging
from pypyr.cache.cache import Cache
import pypyr.moduleloader

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)


class StepCache(Cache):
    """Get functions from the Step cache."""

    def get_step(self, step_name):
        """Get cached run_step function. Adds to cache if not exist.

        Args:
            step_name: load the step module specified by this, get its
                       run_step function and add it to cache.

        Return:
            function: the run_step function of module specified by step_name.
        """
        logger.debug("starting")

        runstep_function = self.get(step_name,
                                    lambda: load_the_step(step_name))

        logger.debug("done")
        return runstep_function


# global instance of the cache. use this to access the cache from elsewhere.
step_cache = StepCache()


def load_the_step(step_name):
    """Load the module specified by step_name, returns its function.

    Args:
        step_name: (string) name of Step module to load

    Returns:
        function: the run_step function in the module
    """
    logger.debug("starting")

    step_module = pypyr.moduleloader.get_module(step_name)
    logger.debug("step module loaded: %s", step_module)

    try:
        run_step_function = getattr(step_module, 'run_step')
    except AttributeError:
        logger.error("The step %s in module %s "
                     "doesn't have a run_step(context) function.",
                     step_name, step_module)
        raise
    logger.debug("done")

    return run_step_function



"""pypyr step that loads json file into context."""
from collections.abc import MutableMapping
import json
import logging
from pypyr.utils.asserts import assert_key_has_value
# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Load a json file into the pypyr context.

    json parsed from the file will be merged into the pypyr context. This will
    overwrite existing values if the same keys are already in there.
    I.e if file json has {'eggs' : 'boiled'} and context {'eggs': 'fried'}
    already exists, returned context['eggs'] will be 'boiled'.

    The json should not be an array [] on the top level, but rather an Object.

    Args:
        context: pypyr.context.Context. Mandatory.
                 The following context key must exist
                - fetchJson
                    - path. path-like. Path to file on disk.
                    - key. string. If exists, write json structure to this
                      context key. Else json writes to context root.

    Also supports a passing path as string to fetchJson, but in this case you
    won't be able to specify a key.

    All inputs support formatting expressions.

    Returns:
        None. updates context arg.

    Raises:
        FileNotFoundError: take a guess
        pypyr.errors.KeyNotInContextError: fetchJson.path missing in context.
        pypyr.errors.KeyInContextHasNoValueError: fetchJson.path exists but is
                                                  None.

    """
    logger.debug("started")

    context.assert_key_has_value(key='fetchJson', caller=__name__)

    fetch_json_input = context.get_formatted('fetchJson')

    if isinstance(fetch_json_input, str):
        file_path = fetch_json_input
        destination_key = None
    else:
        assert_key_has_value(obj=fetch_json_input,
                             key='path',
                             caller=__name__,
                             parent='fetchJson')
        file_path = fetch_json_input['path']
        destination_key = fetch_json_input.get('key', None)

    logger.debug("attempting to open file: %s", file_path)
    with open(file_path) as json_file:
        payload = json.load(json_file)

    if destination_key:
        logger.debug("json file loaded. Writing to context %s",
                     destination_key)
        context[destination_key] = payload
    else:
        if not isinstance(payload, MutableMapping):
            raise TypeError(
                'json input should describe an object at the top '
                'level when fetchJsonKey isn\'t specified. You should have '
                'something like {"key1": "value1", "key2": "value2"} '
                'in the json top-level, not ["value1", "value2"]')

        logger.debug("json file loaded. Merging into pypyr context. . .")
        context.update(payload)

    logger.info("json file written into pypyr context. Count: %s",
                len(payload))
    logger.debug("done")
